

# Libs
import numpy as np
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import logging
logging.getLogger('tensorflow').setLevel(logging.FATAL)
import tensorflow.compat.v1 as tf
import shutil
import skimage.io
import glob
from skimage import io as skio
import pkgutil

tf.disable_v2_behavior()


# from other .py scripts
if pkgutil.find_loader("UNet") is not None:
    from toolbox.imtools import tifread, normalize, im2double, imwrite
    from toolbox.ftools import pathjoin, saveData, loadData
    from toolbox.PartitionOfImage import PI2D
    from toolbox.GPUselect import pick_gpu_lowest_memory
else:
    from .toolbox.imtools import tifread, normalize, im2double, imwrite
    from .toolbox.ftools import pathjoin, saveData, loadData
    from .toolbox.PartitionOfImage import PI2D
    from .toolbox.GPUselect import pick_gpu_lowest_memory

#import pathlib
#from scipy import misc
#from skimage.transform import rotate
#import argparse
#import tifffile
#import sys, math
#import scipy.io as sio
#import fnmatch, glob
#import skimage.exposure as sk
#from tensorflow import keras
#from tensorflow.keras import layers
#from toolbox.imtools import *
#from toolbox.ftools import *
#from toolbox.PartitionOfImage import PI2D
#from toolbox import GPUselect
#from .toolbox.imtools import tifread, tifwrite, imshow, imshowlist, imread, imwrite, im2double, size, imresizeDouble, imresize3Double, imresizeUInt8, imresize3UInt8, normalize, snormalize, cat, imerode, imdilate, imerode3, imdilate3, sphericalStructuralElement, medfilt, maxfilt, minfilt, ptlfilt, imgaussfilt, imlogfilt, imgradmag, localstats, localstats3, imderivatives, imderivatives3, imfeatures, imfeatures3, stack2list, thrsegment
#from .toolbox.ftools import fileparts, listfiles, listsubdirs, pathjoin, saveData, loadData, createFolderIfNonExistent, moveFile, copyFile, removeFile



# Functions 
def concat3 (lst):
    return tf.concat(lst, 3)


class UNet2D:
	hp = None  # hyper-parameters
	nn = None  # network
	tfTraining = None  # if training or not (to handle batch norm)
	tfData = None  # data placeholder
	Session = None
	DatasetMean = 0
	DatasetStDev = 0

	def setupWithHP(hp):
		UNet2D.setup(hp['imSize'],
					 hp['nChannels'],
					 hp['nClasses'],
					 hp['nOut0'],
					 hp['featMapsFact'],
					 hp['downSampFact'],
					 hp['ks'],
					 hp['nExtraConvs'],
					 hp['stdDev0'],
					 hp['nLayers'],
					 hp['batchSize'])

	def setup(imSize, nChannels, nClasses, nOut0, featMapsFact, downSampFact, kernelSize, nExtraConvs, stdDev0,
			  nDownSampLayers, batchSize):
		UNet2D.hp = {'imSize': imSize,
					 'nClasses': nClasses,
					 'nChannels': nChannels,
					 'nExtraConvs': nExtraConvs,
					 'nLayers': nDownSampLayers,
					 'featMapsFact': featMapsFact,
					 'downSampFact': downSampFact,
					 'ks': kernelSize,
					 'nOut0': nOut0,
					 'stdDev0': stdDev0,
					 'batchSize': batchSize}
		tf.reset_default_graph()
		nOutX = [UNet2D.hp['nChannels'], UNet2D.hp['nOut0']]
		dsfX = []
		for i in range(UNet2D.hp['nLayers']):
			nOutX.append(nOutX[-1] * UNet2D.hp['featMapsFact'])
			dsfX.append(UNet2D.hp['downSampFact'])

		# --------------------------------------------------
		# downsampling layer
		# --------------------------------------------------
		with tf.name_scope('placeholders'):
			UNet2D.tfTraining = tf.placeholder(tf.bool, name='training')
			UNet2D.tfData = tf.placeholder("float", shape=[None, UNet2D.hp['imSize'], UNet2D.hp['imSize'],
														   UNet2D.hp['nChannels']], name='data')

		def down_samp_layer(data, index):
			regularizer = tf.keras.regularizers.l1(0.0000000012)
			with tf.variable_scope('ld%d' % index):
				ldXWeights1 = tf.Variable(
					tf.truncated_normal([UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index], nOutX[index + 1]],
										stddev=stdDev0), name='kernelD%d' % index)
				# ldXWeights1 = tf.get_variable(
				# 	initializer=tf.contrib.layers.variance_scaling_initializer(mode='FAN_IN'),
				# 	shape=[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index], nOutX[index + 1]], name='kernelD%d' % index, regularizer=regularizer)
				ldXWeightsExtra = []
				for i in range(nExtraConvs):
					ldXWeightsExtra.append(
						tf.get_variable(initializer=tf.compat.v1.keras.initializers.VarianceScaling(mode='fan_in'),
										shape=[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index + 1], nOutX[index + 1]],
										name='kernelExtra%d' % i))
					# ldXWeightsExtra.append(tf.Variable(
					# 	tf.truncated_normal([UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index + 1], nOutX[index + 1]],
					# 						stddev=stdDev0), name='kernelExtra%d' % i))

				c00 = tf.nn.conv2d(data, ldXWeights1, strides=[1, 1, 1, 1], padding='SAME')
				for i in range(nExtraConvs):
					c00 = tf.nn.conv2d(tf.nn.leaky_relu(c00), ldXWeightsExtra[i], strides=[1, 1, 1, 1], padding='SAME')

				ldXWeightsShortcut = tf.get_variable(initializer=tf.compat.v1.keras.initializers.VarianceScaling(mode='fan_in'),
													 shape=[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index],
															nOutX[index + 1]],
													 name='shortcutWeights', regularizer=regularizer)
				# ldXWeightsShortcut = tf.Variable(
				# 	tf.truncated_normal([1, 1, nOutX[index], nOutX[index + 1]], stddev=stdDev0), name='shortcutWeights')
				shortcut = tf.nn.conv2d(data, ldXWeightsShortcut, strides=[1, 1, 1, 1], padding='SAME')

				bn = tf.nn.leaky_relu(tf.layers.batch_normalization(c00+shortcut, training=UNet2D.tfTraining))
				# bn = tf.layers.batch_normalization(tf.nn.leaky_relu(c00 + shortcut), training=UNet2D.tfTraining)
				# bn = tf.layers.dropout(bn, 0.05 * index, training=UNet2D.tfTraining)
				return tf.nn.max_pool(bn, ksize=[1, dsfX[index], dsfX[index], 1],
									  strides=[1, dsfX[index], dsfX[index], 1], padding='SAME', name='maxpool')

		# --------------------------------------------------
		# bottom layer
		# --------------------------------------------------

		with tf.variable_scope('lb'):
			regularizer = tf.keras.regularizers.l1(0.0000000012)#%0.00012
			lbWeights1 = tf.get_variable(initializer=tf.compat.v1.keras.initializers.VarianceScaling(mode='fan_in'),
										 shape=[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[UNet2D.hp['nLayers']],
												nOutX[UNet2D.hp['nLayers'] + 1]],
										 name='kernel1', regularizer=regularizer)
			# lbWeights1 = tf.Variable(tf.truncated_normal(
			# 	[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[UNet2D.hp['nLayers']], nOutX[UNet2D.hp['nLayers'] + 1]],
			# 	stddev=stdDev0), name='kernel1')
			def lb(hidden):
				# lbn= tf.nn.leaky_relu(
				# 	tf.nn.conv2d(hidden, lbWeights1, strides=[1, 1, 1, 1], padding='SAME'), name='conv')
				lbn = tf.nn.leaky_relu(tf.layers.batch_normalization(
					tf.nn.conv2d(hidden, lbWeights1, strides=[1, 1, 1, 1], padding='SAME'),
					name='conv',training=UNet2D.tfTraining))
				return tf.layers.dropout(lbn, 0.15, training=UNet2D.tfTraining)

		# --------------------------------------------------
		# downsampling
		# --------------------------------------------------

		with tf.name_scope('downsampling'):
			dsX = []
			dsX.append(UNet2D.tfData)

			for i in range(UNet2D.hp['nLayers']):
				dsX.append(down_samp_layer(dsX[i], i))

			b = lb(dsX[UNet2D.hp['nLayers']])

		# --------------------------------------------------
		# upsampling layer
		# --------------------------------------------------
        


		def up_samp_layer(data, index):
			with tf.variable_scope('lu%d' % index):
				regularizer = tf.keras.regularizers.l1(0.0000000012)
				luXWeights1 = tf.get_variable(
					initializer=tf.compat.v1.keras.initializers.VarianceScaling(mode='fan_in'),
					shape=[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index + 1], nOutX[index + 2]],
					name='kernelU%d' % index,regularizer=regularizer)
				luXWeights2 = tf.get_variable(
					initializer=tf.compat.v1.keras.initializers.VarianceScaling(mode='fan_in'),
					shape=[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index] + nOutX[index + 1], nOutX[index + 1]],
					name='kernel2',regularizer=regularizer)
				luXWeightsExtra = []
				for i in range(nExtraConvs):
					luXWeightsExtra.append(tf.get_variable(
						initializer=tf.compat.v1.keras.initializers.VarianceScaling(mode='fan_in'),
						shape=[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index + 1], nOutX[index + 1]],
						name='kernel2Extra%d' % i))
				# luXWeights1 = tf.Variable(
				# 	tf.truncated_normal([UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index + 1], nOutX[index + 2]],
				# 						stddev=stdDev0), name='kernel1')
				# luXWeights2 = tf.Variable(tf.truncated_normal(
				# 	[UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index] + nOutX[index + 1], nOutX[index + 1]],
				# 	stddev=stdDev0), name='kernel2')
				# luXWeightsExtra = []
				# for i in range(nExtraConvs):
				# 	luXWeightsExtra.append(tf.Variable(
				# 		tf.truncated_normal([UNet2D.hp['ks'], UNet2D.hp['ks'], nOutX[index + 1], nOutX[index + 1]],
				# 							stddev=stdDev0), name='kernel2Extra%d' % i))

				outSize = UNet2D.hp['imSize']
				for i in range(index):
					outSize /= dsfX[i]
				outSize = int(outSize)

				outputShape = [UNet2D.hp['batchSize'], outSize, outSize, nOutX[index + 1]]
				us = tf.nn.leaky_relu(
					tf.nn.conv2d_transpose(data, luXWeights1, outputShape, strides=[1, dsfX[index], dsfX[index], 1],
										   padding='SAME'), name='conv1')
				cc = concat3([dsX[index], us])
				# cv = tf.nn.leaky_relu(tf.nn.conv2d(cc, luXWeights2, strides=[1, 1, 1, 1], padding='SAME'), name='conv2')
				cv = tf.nn.leaky_relu(
					tf.layers.batch_normalization(tf.nn.conv2d(cc, luXWeights2, strides=[1, 1, 1, 1], padding='SAME'),
												  name='conv2',
												  training=UNet2D.tfTraining))
				for i in range(nExtraConvs):
					cv = tf.nn.leaky_relu(tf.nn.conv2d(cv, luXWeightsExtra[i], strides=[1, 1, 1, 1], padding='SAME'),
										  name='conv2Extra%d' % i)
				# cv = tf.layers.dropout(cv, 0.25 - 0.05 * index, training=UNet2D.tfTraining)
				return cv

		# --------------------------------------------------
		# final (top) layer
		# --------------------------------------------------

		with tf.variable_scope('lt'):
			regularizer = tf.keras.regularizers.l1(0.0000000012)
			# ltWeights1 = tf.Variable(tf.truncated_normal([1, 1, nOutX[1], nClasses], stddev=stdDev0), name='kernel')
			ltWeights1 = tf.get_variable(initializer=tf.compat.v1.keras.initializers.VarianceScaling(mode='fan_in'),
										 shape=[1, 1, nOutX[1], nClasses],
										 name='kernel',regularizer=regularizer)
			def lt(hidden):
				# return tf.nn.conv2d(hidden, ltWeights1, strides=[1, 1, 1, 1], padding='SAME', name='conv')
				return tf.layers.batch_normalization(
					tf.nn.conv2d(hidden, ltWeights1, strides=[1, 1, 1, 1], padding='SAME', name='conv'),
					training=UNet2D.tfTraining)
		# --------------------------------------------------
		# upsampling
		# --------------------------------------------------

		with tf.name_scope('upsampling'):
			usX = []
			usX.append(b)

			for i in range(UNet2D.hp['nLayers']):
				usX.append(up_samp_layer(usX[i], UNet2D.hp['nLayers'] - 1 - i))

			t = lt(usX[UNet2D.hp['nLayers']])

		sm = tf.nn.softmax(t, -1)
		UNet2D.nn = sm


	def train(trainPath, validPath, testPath, artTrainPath, artValidPath, logPath, modelPath, pmPath, restoreVariables, nSteps, gpuIndex,
			  testPMIndex):
		os.environ['CUDA_VISIBLE_DEVICES'] = '%d' % gpuIndex
		print(trainPath)
		outLogPath = logPath
		trainWriterPath = pathjoin(logPath, 'Train')
		validWriterPath = pathjoin(logPath, 'Valid')
		outModelPath = pathjoin(modelPath, 'model.ckpt')
		outPMPath = pmPath

		batchSize = UNet2D.hp['batchSize']
		imSize = UNet2D.hp['imSize']
		nChannels = UNet2D.hp['nChannels']
		nClasses = UNet2D.hp['nClasses']


		trainingFiles = glob.glob(str(trainPath) + '/' + '*_img.tif')
		trainingMasks = glob.glob(str(trainPath) + '/' +'*_mask.tif')
		artTrainFiles = glob.glob(str(artTrainPath)  + '/' +'*_img.tif')
		artTrainMasks = glob.glob(str(artTrainPath) + '/' +'*_mask.tif')
		trainingFiles = trainingFiles + artTrainFiles
		trainingMasks = trainingMasks + artTrainMasks
		nTrain = len(trainingFiles)
		nSteps = round(nTrain / batchSize) * nSteps  #convert epochs to steps
		print('Training for ' + str(nSteps) + ' steps')

		validFiles = glob.glob(str(validPath) + '/' + '*_img.tif')
		validMasks = glob.glob(str(validPath) + '/' + '*_mask.tif')
		artValidFiles = glob.glob(str(artValidPath) + '/' + '*_img.tif')
		artValidMasks = glob.glob(str(artValidPath) + '/' + '*_mask.tif')
		validFiles = validFiles + artValidFiles
		validMasks = validMasks + artValidMasks
		nValid = len(validFiles)

		testFiles = glob.glob(str(testPath) + '/' + '*_img.tif')
		testMasks = glob.glob(str(testPath) +'/' + '*_mask.tif')
		nTest = len(testFiles)

		nArtTrain = len(artTrainFiles)
		nArtValid = len(artValidFiles)

		print('Found ' + str(nTrain) + ' training images')
		print('Found ' + str(nValid) + ' validation images')
		print('Found ' + str(nTest) + ' test images')
		print('Of these, ' + str(nArtTrain) + ' are artefact training images')
		print(' and  ' + str(nArtValid) + ' artefact validation images')

		# --------------------------------------------------
		# data
		# --------------------------------------------------
		nAug =12
		Train = np.zeros((nTrain, imSize, imSize, nChannels))
		Valid = np.zeros((nValid, imSize, imSize, nChannels))
		Test = np.zeros((nTest, imSize, imSize, nChannels))
		LTrain = np.zeros((nTrain, imSize, imSize, nClasses))
		LValid = np.zeros((nValid, imSize, imSize, nClasses))
		LTest = np.zeros((nTest, imSize, imSize, nClasses))
		WTrain = np.zeros((nTrain, imSize, imSize, nClasses))
		WValid = np.zeros((nValid, imSize, imSize, nClasses))
		WTest = np.zeros((nTest, imSize, imSize, nClasses))

		if not os.path.exists(modelPath):
			os.makedirs(modelPath)
		# if restoreVariables:
		# 	datasetMean = loadData(pathjoin(modelPath,'datasetMean.data'))
		# 	datasetStDev = loadData(pathjoin(modelPath,'datasetStDev.data'))
		# else:
		datasetMean = 0
		datasetStDev = 0
		for iSample in range(0, nTest):
			path = testFiles[iSample]#'%s/I%05d_Img.tif' % (testPath, iSample)
			im = skio.imread(path)
			datasetMean += np.mean(im)
			datasetStDev += np.std(im)

		# datasetMean /= (nTest)
		# datasetStDev /= (nTest)
		datasetMean = 0
		datasetStDev = 1

		print('Using '+ str(datasetMean) + ' and ' + str(datasetStDev) + ' for mean and standard deviation.')
		saveData(datasetMean, pathjoin(modelPath, 'datasetMean.data'))
		saveData(datasetStDev, pathjoin(modelPath, 'datasetStDev.data'))

		fgPixels=0
		bgPixels=0
		for iSample in range(0, nTest):
			path = testFiles[iSample]#'%s/I%05d_Img.tif' % (testPath, iSample)
			for iChan in range(nChannels):
				im = skio.imread(path)
				Test[iSample, :, :, iChan] = im # (im - datasetMean) / datasetStDev

			path = testMasks[iSample]#'%s/I%05d_Ant.png' % (testPath, iSample)
			im = tifread(path)
			fgPixels = fgPixels + np.sum(im == 2)
			bgPixels = bgPixels + np.sum(im == 1)
			path = '%s/I%05d_wt.tif' % (testPath, iSample)
			# W = tifread(path)
			for i in range(nClasses):
				LTest[iSample, :, :, i] = (im == i + 1)
		maxGlobalInt = np.percentile(Test, 99)
		minGlobalInt = np.amin(Test)
		print('Using ' + str(maxGlobalInt) + ' and ' + str(minGlobalInt) + ' for global max and min intensities.')

		bgWeight = 1
		nucleiWeight = bgPixels/fgPixels
		print('Class balance ratio is ' + str(nucleiWeight))

		# Train Loader
		perm = np.arange(nTrain)
		np.random.shuffle(perm)
		for iSample in range(0, nTrain):
			path = trainingFiles[iSample]#'%s/I%05d_Img.tif' % (imPath, perm[iSample])
			for iChan in range(nChannels):
				im = skio.imread(path)
				Train[iSample, :, :, iChan] = im# (im - datasetMean) / datasetStDev
			path = trainingMasks[iSample]#'%s/I%05d_Ant.png' % (imPath, perm[iSample])
			im = tifread(path)
			W = tifread(path)
			for i in range(nClasses):
				LTrain[iSample, :, :, i] = (im == i + 1)
				if i == 0:
					WTrain[iSample, :, :, i] = (W > 0)
				if i == 1:
					WTrain[iSample, :, :, i] = WTrain[iSample, :, :, i] + (W == i + 1) * nucleiWeight

		# Valid Loader
		permV = np.arange(nValid)
		np.random.shuffle(permV)
		for iSample in range(0, nValid):
			path = validFiles[iSample]#'%s/I%05d_Img.tif' % (validPath, permV[iSample])
			# im = im2double(tifread(path))
			for iChan in range(nChannels):
				im = skio.imread(path)
				Valid[iSample, :, :, iChan] = im# (im - datasetMean) / datasetStDev
			path = validMasks[iSample]#'%s/I%05d_Ant.png' % (validPath, permV[iSample])
			im = tifread(path)
			W = tifread(path)
			for i in range(nClasses):
				LValid[iSample, :, :, i] = (im == i + 1)
				if i == 0:
					WValid[iSample, :, :, i] = (W > 0)
				if i == 1:
					WValid[iSample, :, :, i] = WValid[iSample, :, :, i] + (W == i + 1) * nucleiWeight



		# --------------------------------------------------
		# optimization
		# --------------------------------------------------

		tfLabels = tf.placeholder("float", shape=[None, imSize, imSize, nClasses], name='labels')
		tfWeights = tf.placeholder("float", shape=[None, imSize, imSize, nClasses], name='weights')
		globalStep = tf.Variable(0, trainable=False)
		learningRate0 = 0.0005
		decaySteps = 5000
		decayRate = 0.98
		learningRate = tf.train.exponential_decay(learningRate0, globalStep, decaySteps, decayRate, staircase=True)

		with tf.name_scope('optim'):
			l2_loss = tf.losses.get_regularization_loss()
			eps = 1e-7
			log_p = tf.log(tf.clip_by_value(UNet2D.nn, eps, 1 - eps))
			loss = tf.reduce_mean(-tf.reduce_sum(tf.multiply(tf.cast(tfWeights, tf.float32),
															 tf.multiply(tf.cast(tfLabels, tf.float32),
																		 log_p)), 3)) + l2_loss


			updateOps = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
			# optimizer = tf.train.MomentumOptimizer(1e-3,0.9)
			# optimizer = tf.train.MomentumOptimizer(learningRate,0.99)
			optimizer = tf.train.AdamOptimizer(learning_rate=learningRate)
			with tf.control_dependencies(updateOps):
				optOp = optimizer.minimize(loss, global_step=globalStep)

		# for g, v in gradients:
		# 	tf.summary.histogram(v.name, v)
		# 	tf.summary.histogram(v.name + '_grad', g)

		with tf.name_scope('eval'):
			error = []
			for iClass in range(nClasses):
				labels0 = tf.reshape(tf.to_int32(tf.slice(tfLabels, [0, 0, 0, iClass], [-1, -1, -1, 1])),
									 [batchSize, imSize, imSize])
				predict0 = tf.reshape(tf.to_int32(tf.equal(tf.argmax(UNet2D.nn, 3), iClass)),
									  [batchSize, imSize, imSize])
				correct = tf.multiply(labels0, predict0)
				nCorrect0 = tf.reduce_sum(correct)
				nLabels0 = tf.reduce_sum(labels0)
				error.append(1 - tf.to_float(nCorrect0) / tf.to_float(nLabels0))
			errors = tf.tuple(error)

		# --------------------------------------------------
		# inspection
		# --------------------------------------------------

		with tf.name_scope('scalars'):
			tf.summary.scalar('avg_cross_entropy', loss)
			for iClass in range(nClasses):
				tf.summary.scalar('avg_pixel_error_%d' % iClass, error[iClass])
			tf.summary.scalar('learning_rate', learningRate)
		with tf.name_scope('images'):
			split0 = tf.slice(UNet2D.nn, [0, 0, 0, 1], [-1, -1, -1, 1])
			split1 = tf.slice(UNet2D.tfData, [0, 0, 0, 0], [-1, -1, -1, 1])

			planeImN = tf.div(tf.subtract(split1, tf.reduce_min(split1, axis=(1, 2), keep_dims=True)),
							  tf.subtract(tf.reduce_max(split1, axis=(1, 2), keep_dims=True),
										  tf.reduce_min(split1, axis=(1, 2), keep_dims=True)))
			# splitL = tf.slice(UNet2D.tfData, [0, 0, 0, 1], [-1, -1, -1, 1])
			# planeImN2 = tf.div(tf.subtract(splitL, tf.reduce_min(splitL, axis=(1, 2), keep_dims=True)),
			# 				  tf.subtract(tf.reduce_max(splitL, axis=(1, 2), keep_dims=True),
			# 							  tf.reduce_min(splitL, axis=(1, 2), keep_dims=True)))

			plane = tf.concat([planeImN, split0], 2)
			# split2 = tf.slice(UNet2D.nn, [0, 0, 0, 1], [-1, -1, -1, 1])

			# planeImN2 = tf.div(tf.subtract(split3, tf.reduce_min(split3, axis=(1, 2), keep_dims=True)),
			#                   tf.subtract(tf.reduce_max(split3, axis=(1, 2), keep_dims=True),
			#                               tf.reduce_min(split3, axis=(1, 2), keep_dims=True)))
			# plane = tf.concat([plane, split2], 2)
			tf.summary.image('impm', plane, max_outputs=4)
		merged = tf.summary.merge_all()

		# --------------------------------------------------
		# session
		# --------------------------------------------------

		saver = tf.train.Saver()
		config = tf.ConfigProto()
		config.gpu_options.allow_growth = True
		config.allow_soft_placement = True
		sess = tf.Session(config=config)  # config parameter needed to save variables when using GPU

		if os.path.exists(outLogPath):
			shutil.rmtree(outLogPath)
		trainWriter = tf.summary.FileWriter(trainWriterPath, sess.graph)
		validWriter = tf.summary.FileWriter(validWriterPath, sess.graph)

		if restoreVariables:
			saver.restore(sess, outModelPath)
			print("Model restored.")
		else:
			sess.run(tf.global_variables_initializer())

		# --------------------------------------------------
		# train
		# --------------------------------------------------

		batchData = np.zeros((batchSize, imSize, imSize, nChannels))
		batchLabels = np.zeros((batchSize, imSize, imSize, nClasses))
		batchWeights = np.zeros((batchSize, imSize, imSize, nClasses))
		permT = np.arange(nTrain)
		np.random.shuffle(permT)

		permV = np.arange(nValid)
		np.random.shuffle(permV)

		maxBrig = 50#1 * datasetStDev
		maxCont = 0.3#0.1 * datasetStDev
		jT = 0
		jV = 0
		epochCounter = 1
		for i in range(nSteps):
			# train

			for j in range(batchSize):
				fBrig = maxBrig * np.float_power(-1, np.random.rand() < 0.5) * np.random.rand()
				fCont = 1 + maxCont * np.float_power(-1, np.random.rand() < 0.5) * np.random.rand()
				image = np.zeros((imSize, imSize, nChannels))
				for iChan in range(nChannels):
					image[:, :, iChan] = Train[permT[jT + j], :, :, iChan] * fCont + fBrig

					augmentType = np.random.rand()
					if augmentType < 0.2: # global normalization
						image = (((image - minGlobalInt) / maxGlobalInt) * 255 - datasetMean) / datasetStDev
					elif augmentType < 0.4: # local normalization
						image = ((image - np.amin(image)) / (
									np.percentile(image, 99.9) - np.amin(image)) * 255 - datasetMean) / datasetStDev
					else:
						image = (image-datasetMean) / datasetStDev

				augmentType = np.random.rand()
				if augmentType < 0.3:
					batchData[j, :, :, :] = skimage.transform.rotate(image,90)
					batchLabels[j, :, :, :] = skimage.transform.rotate(LTrain[permT[jT + j], :, :, :],90)
					batchWeights[j, :, :, :] = skimage.transform.rotate(WTrain[permT[jT + j], :, :, :],90)
				elif augmentType < 0.6:
					batchData[j, :, :, :] = skimage.transform.rotate(image, 180)
					batchLabels[j, :, :, :] = skimage.transform.rotate(LTrain[permT[jT + j],:,:,:], 180)
					batchWeights[j, :, :, :] = skimage.transform.rotate(WTrain[permT[jT + j], :, :, :], 180)
				else:
					batchData[j, :, :, :] = skimage.transform.rotate(image, 0)
					batchLabels[j, :, :, :] = skimage.transform.rotate(LTrain[permT[jT + j],:,:,:], 0)
					batchWeights[j, :, :, :] = skimage.transform.rotate(WTrain[permT[jT + j], :, :, :], 0)

			batchWeights[j, :, :, :] = WTrain[permT[jT + j], :, :, :]
			summary, _ = sess.run([merged, optOp], feed_dict={UNet2D.tfData: batchData, tfLabels: batchLabels,
															  tfWeights: batchWeights, UNet2D.tfTraining: 1})
			jT = jT + batchSize
			if jT > (nTrain - batchSize - 1):
				jT = 0
				np.random.shuffle(permT)
				epochCounter = epochCounter + 1
			if np.mod(i, 20) == 0:
				trainWriter.add_summary(summary, i)

			# validation
			for j in range(batchSize):
				fBrig = maxBrig * np.float_power(-1, np.random.rand() < 0.5) * np.random.rand()
				fCont = 1 + maxCont * np.float_power(-1, np.random.rand() < 0.5) * np.random.rand()
				image = np.zeros((imSize, imSize, nChannels))
				image[:, :, 0] = Valid[permV[jV + j], :, :, 0] * fCont + fBrig
				augmentType = np.random.rand()
				if augmentType < 0.2:  # global normalization
					image = (((image - minGlobalInt) / maxGlobalInt) * 255 - datasetMean) / datasetStDev
				elif augmentType < 0.4:  # local normalization
					image = ((image - np.amin(image)) / (
								np.percentile(image, 99.9) - np.amin(image)) * 255 - datasetMean) / datasetStDev
				else:
					image = (image - datasetMean) / datasetStDev
				batchData[j, :, :, :] = image
				batchLabels[j, :, :, :] = LValid[permV[jV + j], :, :, :]
				batchWeights[j, :, :, :] = WValid[permV[jV + j], :, :, :]
			summary, es = sess.run([merged, errors], feed_dict={UNet2D.tfData: batchData, tfLabels: batchLabels,
																tfWeights: batchWeights, UNet2D.tfTraining: 0})
			jV = jV + batchSize
			if jV > (nValid - batchSize - 1):
				jV = 0
				np.random.shuffle(permV)
			if np.mod(i, 20) == 0:
				validWriter.add_summary(summary, i)

			e = np.mean(es)
			print('step %05d, e: %f' % (i, e) + ', epoch: ' + str(epochCounter))

			if i == 0:
				if restoreVariables:
					lowestError = e
				else:
					lowestError = np.inf

			if np.mod(i, 1000) == 0 and e < lowestError:
				lowestError = e
				print("Model saved in file: %s" % saver.save(sess, outModelPath))

		# --------------------------------------------------
		# save hyper-parameters, clean-up
		# --------------------------------------------------

		saveData(UNet2D.hp, pathjoin(modelPath, 'hp.data'))

		trainWriter.close()
		validWriter.close()
		sess.close()


		# --------------------------------------------------
		# test
		# --------------------------------------------------
		os.environ['CUDA_VISIBLE_DEVICES'] = '%d' % gpuIndex
		tf.reset_default_graph()
		variablesPath = pathjoin(modelPath, 'model.ckpt')
		outPMPath = pmPath

		hp = loadData(pathjoin(modelPath, 'hp.data'))
		UNet2D.setupWithHP(hp)
		saver = tf.train.Saver()
		sess = tf.Session(config=tf.ConfigProto(
			allow_soft_placement=True))  # config parameter needed to save variables when using GPU
		saver.restore(sess, variablesPath)
		print("Model restored.")

		if not os.path.exists(outPMPath):
			os.makedirs(outPMPath)

		for i in range(nTest):
			j = np.mod(i, batchSize)
			image = np.zeros((imSize, imSize, nChannels))
			image[:,:,0] = Test[i, :, :, 0]
			batchData[j, :, :, :] = image
			batchLabels[j, :, :, :] = LTest[i, :, :, :]

			if j == batchSize - 1 or i == nTest - 1:

				output = sess.run(UNet2D.nn,
								  feed_dict={UNet2D.tfData: batchData, UNet2D.tfTraining: 0})

				for k in range(j + 1):
					pm = output[k, :, :, 1]
					gt = batchLabels[k, :, :, 1]
					im = normalize(batchData[k, :, :, 0])
					imwrite(np.uint8(255 * np.concatenate((im, np.concatenate((pm, gt), axis=1)), axis=1)),
							'%s/I%05d_Nuc.png' % (outPMPath, i - j + k + 1))

				for k in range(j + 1):
					pm = output[k, :, :, 0]
					gt = batchLabels[k, :, :, 0]
					im = normalize(batchData[k, :, :, 0])
					imwrite(np.uint8(255 * np.concatenate((im, np.concatenate((pm, gt), axis=1)), axis=1)),
							'%s/I%05d_BG.png' % (outPMPath, i - j + k + 1))

	def deploy(imPath, nImages, modelPath, pmPath, gpuIndex, pmIndex):
		os.environ['CUDA_VISIBLE_DEVICES'] = '%d' % gpuIndex
		tf.reset_default_graph()
		variablesPath = pathjoin(modelPath, 'model.ckpt')
		outPMPath = pmPath

		hp = loadData(pathjoin(modelPath, 'hp.data'))
		UNet2D.setupWithHP(hp)

		batchSize = UNet2D.hp['batchSize']
		imSize = UNet2D.hp['imSize']
		nChannels = UNet2D.hp['nChannels']
		nClasses = UNet2D.hp['nClasses']

		# --------------------------------------------------
		# data
		# --------------------------------------------------

		testFiles = glob.glob(testPath + '*_img.tif')
		nTest = len(testFiles)
		Data = np.zeros((nTest, imSize, imSize, nChannels))
		datasetMean = loadData(pathjoin(modelPath, 'datasetMean.data'))
		datasetStDev = loadData(pathjoin(modelPath, 'datasetStDev.data'))

		for iSample in range(0, nTest):
			for iChan in range(nChannels):
				im = im2double(skio.imread(testFiles[iSample], img_num=iChan))
				Data[iSample, :, :, iChan] = im  # (im - datasetMean) / datasetStDev

		# --------------------------------------------------
		# session
		# --------------------------------------------------

		saver = tf.train.Saver()
		sess = tf.Session(config=tf.ConfigProto(
			allow_soft_placement=True))  # config parameter needed to save variables when using GPU

		saver.restore(sess, variablesPath)
		print("Model restored.")

		# --------------------------------------------------
		# deploy
		# --------------------------------------------------

		batchData = np.zeros((batchSize, imSize, imSize, nChannels))

		if not os.path.exists(outPMPath):
			os.makedirs(outPMPath)

		for i in range(nTest):
			print(i, nTest)

			j = np.mod(i, batchSize)

			batchData[j, :, :, :] = Data[i, :, :, :]

			if j == batchSize - 1 or i == nTest - 1:

				output = sess.run(UNet2D.nn, feed_dict={UNet2D.tfData: batchData, UNet2D.tfTraining: 0})

				for k in range(j + 1):
					pm = output[k, :, :, pmIndex]
					im = normalize(batchData[k, :, :, 0])
					# imwrite(np.uint8(255*np.concatenate((im,pm),axis=1)),'%s/I%05d.png' % (outPMPath,i-j+k+1))
					imwrite(np.uint8(255 * im), '%s/I%05d_Im.png' % (outPMPath, i - j + k + 1))
					imwrite(np.uint8(255 * pm), '%s/I%05d_PM.png' % (outPMPath, i - j + k + 1))

		# --------------------------------------------------
		# clean-up
		# --------------------------------------------------

		sess.close()

	def singleImageInferenceSetup(modelPath, gpuIndex, mean, std):

		variablesPath = pathjoin(modelPath, 'model.ckpt')

		hp = loadData(pathjoin(modelPath, 'hp.data'))
		UNet2D.setupWithHP(hp)
		if mean == -1:
			UNet2D.DatasetMean = loadData(pathjoin(modelPath, 'datasetMean.data'))
		else:
			UNet2D.DatasetMean = mean

		if std == -1:
			UNet2D.DatasetStDev = loadData(pathjoin(modelPath, 'datasetStDev.data'))
		else:
			UNet2D.DatasetStDev = std


		# --------------------------------------------------
		# session
		# --------------------------------------------------

		saver = tf.train.Saver()
		UNet2D.Session = tf.Session(config=tf.ConfigProto(
			allow_soft_placement=True))  # config parameter needed to save variables when using GPU

		saver.restore(UNet2D.Session, variablesPath)
		print("Model restored.")

	def singleImageInferenceCleanup():
		UNet2D.Session.close()

	def singleImageInference(image, mode, pmIndex):
		print('Inference...')

		batchSize = UNet2D.hp['batchSize']
		imSize = UNet2D.hp['imSize']
		nChannels = UNet2D.hp['nChannels']

		PI2D.setup(image, imSize, int(imSize / 8), mode)
		PI2D.createOutput(1)

		batchData = np.zeros((batchSize, imSize, imSize, nChannels))
		for i in range(PI2D.NumPatches):
			j = np.mod(i, batchSize)
			test = PI2D.getPatch(i)
			# test = im2double(sk.rescale_intensity(test, in_range=(np.min(test), np.percentile(test,99)), out_range=(0, 255)))
			P = (test - UNet2D.DatasetMean) / UNet2D.DatasetStDev
			for iChan in range(nChannels):
				batchData[j, :, :, iChan] = P  # [iChan, :, :]
			if j == batchSize - 1 or i == PI2D.NumPatches - 1:
				output = UNet2D.Session.run(UNet2D.nn,
											feed_dict={UNet2D.tfData: batchData, UNet2D.tfTraining: 0})
				for k in range(j + 1):
					pm = output[k, :, :, pmIndex]
					PI2D.patchOutput(i - j + k, pm)
		# PI2D.patchOutput(i-j+k,normalize(imgradmag(PI2D.getPatch(i-j+k),1)))

		return PI2D.getValidOutput()




# -*- coding: utf-8 -*-
#Created on Thu Nov 10 09:53:28 2022
#@author: Ajit Johnson Nirmal
#Function to incorporate the feature matrix and probability matrix into an adata object

"""
!!! abstract "Short Description"
    The `csObject` function creates a CSPOT object using the anndata 
    framework by inputting csScore and a pre-calculated single-cell spatial table. 
    This centralizes all information into one file, streamlining the data analysis 
    process and reducing the risk of losing data.

## Function
"""

# libs
import anndata as ad
import pandas as pd
import pathlib
import numpy as np
import os
import argparse

# function
def csObject (spatialTablePath,
                 csScorePath,
                 CellId='CellID',
                 uniqueCellId=True,
                 split='X_centroid',
                 removeDNA=True,
                 remove_string_from_name=None,
                 log=True,
                 dropMarkers=None,
                 verbose=True,
                 projectDir=None):
    """
Parameters:
    spatialTablePath (list):
        Provide a list of paths to the single-cell spatial feature tables, ensuring each image has a unique path specified.

    csScorePath (list):
        Supply a list of paths to the DL score tables created using generateCSScore,
        ensuring they correspond to the image paths specified.

    CellId (str, optional):
        Specify the column name that holds the cell ID (a unique name given to each cell).

    uniqueCellId (bool, optional):
        The function generates a unique name for each cell by combining the CellId and imageid.
        If you don't want this, pass False. In such case the function will default to using just the CellId.
        However, make sure CellId is unique especially when loading multiple images together.

    split (string, optional):
        The spatial feature table generally includes single cell expression data
        and meta data such as X, Y coordinates, and cell shape size. The CSPOT
        object separates them. Ensure that the expression data columns come first,
        followed by meta data columns. Provide the column name that marks the split,
        i.e the column name immediately following the expression data.

    removeDNA (bool, optional):
        Exclude DNA channels from the final output. The function searches for
        column names containing the string `dna` or `dapi`. 

    remove_string_from_name (string, optional):
        Cleans up channel names by removing user specified string from all marker
        names. 

    log (bool, optional):
        Apply log1p transformation to log the data. 

    dropMarkers (list, optional):
        Specify a list of markers to be removed from the analysis, for
        example: ["background_channel", "CD20"]. 

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  

    projectDir (string, optional):
        Provide the path to the output directory. The result will be located at
        `projectDir/CSPOT/csObject/`.

Returns:
    csObject (anndata):
        If projectDir is provided the CSPOT Object will be saved as a
        `.h5ad` file in the provided directory.

Example:
        ```python
        
        # set the working directory & set paths to the example data
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        # Path to all the files that are necessary files for running csObject function
        segmentationPath = projectDir + '/segmentation/exampleSegmentationMask.tif'
        csScorePath = projectDir + '/CSPOT/csScore/exampleImage_cspotPredict.ome.csv'
        
        # please note that there are a number of defaults in the below function that assumes certain structure within the spatialTable.
        # Please confirm it is similar with user data or modifiy the parameters accordingly
        # check out the documentation for further details
        adata = cs.csObject (spatialTablePath=spatialTablePath,
                        csScorePath=csScorePath,
                        CellId='CellID',
                        uniqueCellId=True,
                        split='X_centroid',
                        removeDNA=True,
                        remove_string_from_name=None,
                        log=True,
                        dropMarkers=None,
                        projectDir=projectDir)
        
        # Same function if the user wants to run it via Command Line Interface
        python csObject.py \
            --spatialTablePath /Users/aj/Documents/cspotExampleData/quantification/exampleSpatialTable.csv \
            --csScorePath /Users/aj/Documents/cspotExampleData/CSPOT/csScore/exampleImage_cspotPredict.ome.csv \
            --projectDir /Users/aj/Documents/cspotExampleData
        
        ```

    """

    # spatialTablePath list or string
    if isinstance(spatialTablePath, str):
        spatialTablePath = [spatialTablePath]
    spatialTablePath = [pathlib.Path(p) for p in spatialTablePath]
    # csScorePath list or string
    if isinstance(csScorePath, str):
        csScorePath = [csScorePath]
    csScorePath = [pathlib.Path(p) for p in csScorePath]

    # Import spatialTablePath
    def load_process_data (image):
        # Print the data that is being processed
        if verbose is True:
            print(f"Loading {image.name}")
        d = pd.read_csv(image)
        # If the data does not have a unique image ID column, add one.
        if 'imageid' not in d.columns:
            imid = image.stem
            d['imageid'] = imid
        # Unique name for the data
        if uniqueCellId is True:
            d.index = d['imageid'].astype(str)+'_'+d[CellId].astype(str)
        else:
            d.index = d[CellId]

        # move image id and cellID column to end
        cellid_col = [col for col in d.columns if col != CellId] + [CellId]; d = d[cellid_col]
        imageid_col = [col for col in d.columns if col != 'imageid'] + ['imageid']; d = d[imageid_col]
        # If there is INF replace with zero
        d = d.replace([np.inf, -np.inf], 0)
        # Return data
        return d

    # Import csScorePath
    def load_process_probTable (image):
        d = pd.read_csv(image, index_col=0)
        # Return data
        return d

    # Apply function to all spatialTablePath and create a master dataframe
    r_load_process_data = lambda x: load_process_data(image=x) # Create lamda function
    all_spatialTable = list(map(r_load_process_data, list(spatialTablePath))) # Apply function
    # Merge all the spatialTablePath into a single large dataframe
    for i in range(len(all_spatialTable)):
        all_spatialTable[i].columns = all_spatialTable[0].columns
    entire_spatialTable = pd.concat(all_spatialTable, axis=0, sort=False)

    # Apply function to all csScorePath and create a master dataframe
    r_load_process_probTable = lambda x: load_process_probTable(image=x) # Create lamda function
    all_probTable = list(map(r_load_process_probTable, list(csScorePath))) # Apply function
    # Merge all the csScorePath into a single large dataframe
    for i in range(len(all_probTable)):
        all_probTable[i].columns = all_probTable[0].columns
    entire_probTable = pd.concat(all_probTable, axis=0, sort=False)
    # make the index of entire_probTable same as all_probTable
    ## NOTE THIS IS A HARD COPY WITHOUT ANY CHECKS! ASSUMES BOTH ARE IN SAME ORDER
    entire_probTable.index = entire_spatialTable.index


    # Split the data into expression data and meta data
    # Step-1 (Find the index of the column with name X_centroid)
    split_idx = entire_spatialTable.columns.get_loc(split)
    meta = entire_spatialTable.iloc [:,split_idx:]
    # Step-2 (select only the expression values)
    entire_spatialTable = entire_spatialTable.iloc [:,:split_idx]

    # Rename the columns of the data
    if remove_string_from_name is not None:
        entire_spatialTable.columns = entire_spatialTable.columns.str.replace(remove_string_from_name, '')

    # Save a copy of the column names in the uns space of ANNDATA
    markers = list(entire_spatialTable.columns)

    # Remove DNA channels
    if removeDNA is True:
        entire_spatialTable = entire_spatialTable.loc[:,~entire_spatialTable.columns.str.contains('dna', case=False)]
        entire_spatialTable = entire_spatialTable.loc[:,~entire_spatialTable.columns.str.contains('dapi', case=False)]

    # Drop unnecessary markers
    if dropMarkers is not None:
        if isinstance(dropMarkers, str):
            dropMarkers = [dropMarkers]
        dropMarkers = list(set(dropMarkers).intersection(entire_spatialTable.columns))
        entire_spatialTable = entire_spatialTable.drop(columns=dropMarkers)

    # Create an anndata object
    adata = ad.AnnData(entire_spatialTable, dtype=np.float64)
    adata.obs = meta
    adata.uns['all_markers'] = markers
    adata.uns['csScore'] = entire_probTable

    # Add log data
    if log is True:
        adata.raw = adata
        adata.X = np.log1p(adata.X)

    # Save data if requested
    if projectDir is not None:
        finalPath = pathlib.Path(projectDir + '/CSPOT/csObject')
        if not os.path.exists(finalPath):
            os.makedirs(finalPath)
        if len(spatialTablePath) > 1:
            imid = 'csObject'
        else:
            imid = csScorePath[0].stem
        adata.write(finalPath / f'{imid}.h5ad')
        # Finish Job
        if verbose is True:
            print('CSPOT Object has been created, head over to'+ str(projectDir) + '/CSPOT/csObject" to view results')
    else:
        # Return data
        if verbose is True:
            print('CSPOT Object has been created')
        return adata
    
        

# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Create a csObject from single-cell spatial feature tables.')
    parser.add_argument('--spatialTablePath', type=str, nargs='+', help='Provide a list of paths to the single-cell spatial feature tables.')
    parser.add_argument('--csScorePath', type=str, nargs='+', help='Supply a list of paths to the DL score tables created using generateCSScore.')
    parser.add_argument('--CellId', type=str, default='CellID', help='Specify the column name that holds the cell ID.')
    parser.add_argument('--uniqueCellId', type=bool, default=True, help='The function generates a unique name for each cell by combining the CellId and imageid.')
    parser.add_argument('--split', type=str, default='X_centroid', help='Provide the column name that marks the split between expression data and meta data.')
    parser.add_argument('--removeDNA', type=bool, default=True, help='Exclude DNA channels from the final output.')
    parser.add_argument('--remove_string_from_name', type=str, default=None, help='Cleans up channel names by removing user specified string from all marker names.')
    parser.add_argument('--log', type=bool, default=True, help='Apply log1p transformation to log the data.')
    parser.add_argument('--dropMarkers', type=str, nargs='+', default=None, help='Specify a list of markers to be removed from the analysis.')
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")       
    parser.add_argument('--projectDir', type=str, default=None, help='Provide the path to the output directory.')
    args = parser.parse_args()
    csObject(spatialTablePath=args.spatialTablePath,
                csScorePath=args.csScorePath,
                CellId=args.CellId,
                uniqueCellId=args.uniqueCellId,
                split=args.split,
                removeDNA=args.removeDNA,
                remove_string_from_name=args.remove_string_from_name,
                log=args.log,
                dropMarkers=args.dropMarkers,
                verbose=args.verbose,
                projectDir=args.projectDir)



# -*- coding: utf-8 -*-
# Created on Mon Nov  9 21:00:57 2020
# @author: Ajit Johnson Nirmal
"""
!!! abstract "Short Description"
    Users can utilize the `csExport` function to store the contents of the csObject to a `.CSV` file. 
      
    Keep in mind that the presence of multiple intermediate files in the object will result in the production of several CSV files.

## Function
"""

# Import
import pandas as pd
import argparse
import pathlib
import anndata as ad
import os


# Function
def csExport(
    csObject, projectDir, fileName=None, raw=False, CellID="CellID", verbose=True
):
    """
Parameters:
    csObject (anndata):
        Pass the `csObject` loaded into memory or a path to the `csObject` 
        file (.h5ad).
        
    projectDir (str, optional):
        Provide the path to the output directory. The result will be located at
        `projectDir/CSPOT/csExport/`. 

    fileName (str, optional):
        Specify the name of the CSV output file. If you don't provide a file name, t
        he default name `csExport.csv` will be assigned.
        
    raw (bool, optional):
        If `True` raw data will be returned instead of the CSPOT scaled data.
        
    CellId (str, optional):
        Specify the column name that holds the cell ID (a unique name given to each cell).

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  

Returns:
    CSV files (.csv):
        The `.csv` files can be found under `projectDir/CSPOT/csExport/`

Example:
        ```python
        # path to files needed for csExport
        projectDir = '/Users/aj/Documents/cspotExampleData'
        csObject = projectDir + '/CSPOT/csObject/exampleImage_cspotPredict.ome.h5ad'
        
        cs.csExport(csObject,
               projectDir,
               fileName=None,
               raw=False,
               CellID='CellID',
               verbose=True)
        
        # Same function if the user wants to run it via Command Line Interface
        python csExport.py \
            --csObject /Users/aj/Documents/cspotExampleData/CSPOT/cspotOutput/exampleImage_cspotPredict.ome.h5ad \
            --projectDir /Users/aj/Documents/cspotExampleData
        
        ```

    """

    # Load the andata object
    if isinstance(csObject, str):
        if fileName is None:
            imid = pathlib.Path(csObject).stem
        else:
            imid = str(fileName)
        csObject = ad.read(csObject)
    else:
        if fileName is None:
            imid = "csExport"
        else:
            imid = str(fileName)
        csObject = csObject

    # Expression matrix & obs data
    if raw is True:
        data = pd.DataFrame(
            csObject.raw.X, index=csObject.obs.index, columns=csObject.var.index
        )
    else:
        data = pd.DataFrame(
            csObject.X, index=csObject.obs.index, columns=csObject.var.index
        )
    meta = pd.DataFrame(csObject.obs)
    # Merge the two dataframes
    merged = pd.concat([data, meta], axis=1, sort=False)

    # Add a column to save cell-id
    # make cellID the first column
    if CellID in merged.columns:
        first_column = merged.pop(CellID)
        merged.insert(0, CellID, first_column)
    else:
        merged["CellID"] = merged.index
        first_column = merged.pop(CellID)
        merged.insert(0, CellID, first_column)

    # reset index
    merged = merged.reset_index(drop=True)

    # create a folder to hold the results
    folderPath = pathlib.Path(projectDir + "/CSPOT/csExport/")
    folderPath.mkdir(exist_ok=True, parents=True)

    # extract some of the data stored in .uns and save
    if hasattr(csObject, "uns") and "cspotOutput" in csObject.uns:
        cspotOutput = csObject.uns["cspotOutput"]
        cspotOutput.index = merged["CellID"]
        cspotOutput.to_csv(folderPath / "cspotOutput.csv")
    if hasattr(csObject, "uns") and "csScore" in csObject.uns:
        csScore = csObject.uns["csScore"]
        csScore.index = merged["CellID"]
        csScore.to_csv(folderPath / "csScore.csv")

    # scaled data
    merged.to_csv(folderPath / f"{imid}.csv", index=False)

    # Finish Job
    if verbose is True:
        print(
            'Contents of the csObject have been exported to "'
            + str(projectDir)
            + '/CSPOT/csExport"'
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Export CSPOT results to CSV.")
    parser.add_argument(
        "--csObject", type=str, help="Path to the csObject file (.h5ad)", required=True
    )
    parser.add_argument(
        "--projectDir", type=str, help="Path to the output directory", required=True
    )
    parser.add_argument(
        "--fileName",
        type=str,
        help="Name of the CSV output file",
        default="csExport.csv",
    )
    parser.add_argument(
        "--raw",
        type=bool,
        default=False,
        help="Return raw data instead of CSPOT scaled data",
    )
    parser.add_argument(
        "--CellID",
        type=str,
        help="Column name that holds the cell ID",
        default="CellID",
    )
    parser.add_argument(
        "--verbose",
        type=bool,
        default=True,
        help="Print detailed information about the process",
    )
    args = parser.parse_args()

    csExport(
        csObject=args.csObject,
        projectDir=args.projectDir,
        fileName=args.fileName,
        raw=args.raw,
        CellID=args.CellID,
        verbose=args.verbose,
    )



from os.path import *
from os import listdir, makedirs, remove
import pickle
import shutil

def fileparts(path): # path = file path
    [p,f] = split(path)
    [n,e] = splitext(f)
    return [p,n,e]

def listfiles(path,token): # path = folder path
    l = []
    for f in listdir(path):
        fullPath = join(path,f)
        if isfile(fullPath) and token in f:
            l.append(fullPath)
    l.sort()
    return l

def listsubdirs(path): # path = folder path
    l = []
    for f in listdir(path):
        fullPath = join(path,f)
        if isdir(fullPath):
            l.append(fullPath)
    l.sort()
    return l

def pathjoin(p,ne): # '/path/to/folder', 'name.extension' (or a subfolder)
    return join(p,ne)

def saveData(data,path):
    print('saving data')
    dataFile = open(path, 'wb')
    pickle.dump(data, dataFile)

def loadData(path):
    print('loading data')
    dataFile = open(path, 'rb')
    return pickle.load(dataFile)

def createFolderIfNonExistent(path):
    if not exists(path): # from os.path
        makedirs(path)

def moveFile(fullPathSource,folderPathDestination):
    [p,n,e] = fileparts(fullPathSource)
    shutil.move(fullPathSource,pathjoin(folderPathDestination,n+e))

def copyFile(fullPathSource,folderPathDestination):
    [p,n,e] = fileparts(fullPathSource)
    shutil.copy(fullPathSource,pathjoin(folderPathDestination,n+e))

def removeFile(path):
    remove(path)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Thu Aug 18 16:37:29 2022
#@author: Ajit Johnson Nirmal
#Function to generate Masks for UNET model


"""
!!! abstract "Short Description"
    The function generates a mask for the deep learning model training, using 
    automated approaches. Splitting the data into training, validation and 
    test sets is also included in the function, making it easier to feed the 
    data directly into the deep learning algorithm. Note that manually drawing 
    the mask on thumbnails is the ideal approach, however for scalability 
    purposes, automation is used.


## Function
"""

# Libs
import pathlib
import cv2 as cv
import random
import numpy as np
import tifffile
import argparse

# Function
def generateTrainTestSplit (thumbnailFolder, 
                            projectDir, 
                            file_extension=None,
                            verbose=True,
                            TruePos='TruePos', NegToPos='NegToPos',
                            TrueNeg='TrueNeg', PosToNeg='PosToNeg'):
    """
Parameters:
    thumbnailFolder (list):
        List of folders that contains the human sorted Thumbnails that is to be used
        for generating training data and split them train test and validation cohorts.

    projectDir (str):
        Path to output directory.

    file_extension (str, optional):
        If there are non-image files in the thumbnailFolder, the user can specify
        a file extension to only select those files for processing. The default is None.

    verbose (bool, optional):
        If True, print detailed information about the process to the console. 

    TruePos (str, optional):
        Name of the folder that holds the Thumbnails classified as True Positive.
        The default is 'TruePos'.

    NegToPos (str, optional):
        Name of the folder that holds the Thumbnails classified as True Negative.
        The default is 'NegToPos'.

    TrueNeg (str, optional):
        Name of the folder that holds the Thumbnails that were moved from `True Positive`
        to `True Negative`. The default is 'TrueNeg'.

    PosToNeg (str, optional):
        Name of the folder that holds the Thumbnails that were moved from `True Negative`
        to `True Positive`. The default is 'PosToNeg'.

Returns:
    masks (images):
        Segmentation masks are generated for every Thumbnail and split into Train,
        Test and Validation cohorts.

Example:
        ```python
        
        # High level working directory
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        # Folder where the raw Thumbnails are stored
        thumbnailFolder = [projectDir + '/CSPOT/Thumbnails/CD3D',
                           projectDir + '/CSPOT/Thumbnails/ECAD']
        
        # The function accepts the four pre-defined folders. If you had renamed them, please change it using the parameter below.
        # If you had deleted any of the folders and are not using them, replace the folder name with `None` in the parameter.
        cs.generateTrainTestSplit ( thumbnailFolder, 
                                    projectDir=projectDir,
                                    file_extension=None,
                                    TruePos='TruePos', NegToPos='NegToPos',
                                    TrueNeg='TrueNeg', PosToNeg='PosToNeg')
        
        # Same function if the user wants to run it via Command Line Interface
        python generateTrainTestSplit.py \
            --thumbnailFolder /Users/aj/Desktop/cspotExampleData/CSPOT/Thumbnails/CD3D /Users/aj/Desktop/cspotExampleData/CSPOT/Thumbnails/ECAD \
            --projectDir /Users/aj/Desktop/cspotExampleData/
        
        ```

    """

    # Function takes in path to two folders, processes the images in those folders,
    # and saves them into a different folder that contains Train, Validation and Test samples
    #TruePos='TruePos'; NegToPos='NegToPos'; TrueNeg='TrueNeg'; PosToNeg='PosToNeg'; verbose=True

    # convert the folder into a list
    if isinstance (thumbnailFolder, str):
        thumbnailFolder = [thumbnailFolder]

    # convert all path names to pathlib
    thumbnailFolder = [pathlib.Path(p) for p in thumbnailFolder]
    projectDir = pathlib.Path(projectDir)

    # find all markers passed
    all_markers = [i.stem for i in thumbnailFolder]

    # create directories to save
    for i in all_markers:
        if not (projectDir / 'CSPOT/TrainingData/' / f"{i}" /  'training').exists ():
            (projectDir / 'CSPOT/TrainingData/' / f"{i}" /  'training').mkdir(parents=True, exist_ok=True)

        if not (projectDir / 'CSPOT/TrainingData/' / f"{i}" /  'validation').exists ():
            (projectDir / 'CSPOT/TrainingData/' / f"{i}" /  'validation').mkdir(parents=True, exist_ok=True)

        if not (projectDir / 'CSPOT/TrainingData/' / f"{i}" /  'test').exists ():
            (projectDir / 'CSPOT/TrainingData/' / f"{i}" /  'test').mkdir(parents=True, exist_ok=True)

    # standard format
    if file_extension is None:
        file_extension = '*'
    else:
        file_extension = '*' + str(file_extension)

    # Filter on pos cells
    def pos_filter (path):
        image = cv.imread(str(path.resolve()), cv.IMREAD_GRAYSCALE)
        blur = cv.GaussianBlur(image, ksize=(3,3), sigmaX=1, sigmaY=1)
        ret3,th3 = cv.threshold(blur,0,1,cv.THRESH_OTSU)
        mask = th3 + 1
        return [mask, image]

    # Filter on neg cells
    def neg_filter (path):
        image = cv.imread(str(path.resolve()), cv.IMREAD_GRAYSCALE)
        mask = np.ones(image.shape, dtype=np.uint8)
        return [mask, image]

    # identify the files within all the 4 folders
    def findFiles (folderIndex):
        if verbose is True:
            print ('Processing: ' + str(thumbnailFolder[folderIndex].stem))
        marker_name = str(thumbnailFolder[folderIndex].stem)

        baseFolder = thumbnailFolder[folderIndex]

        if TruePos is not None:
            pos = list(pathlib.Path.glob(baseFolder / TruePos, file_extension))
        if NegToPos is not None:
            negtopos = list(pathlib.Path.glob(baseFolder / NegToPos, file_extension))
        positive_cells = pos + negtopos

        if TrueNeg is not None:
            neg = list(pathlib.Path.glob(baseFolder / TrueNeg, file_extension))
        if PosToNeg is not None:
            postoneg = list(pathlib.Path.glob(baseFolder / PosToNeg, file_extension))
        negative_cells = neg + postoneg

        # prepare the Training, Validataion and Test Cohorts
        if len(positive_cells) > 0:
            train_pos = random.sample(positive_cells, round(len(positive_cells) * 0.6))
            remanining_pos = list(set(positive_cells) - set(train_pos))
            val_pos = random.sample(remanining_pos, round(len(remanining_pos) * 0.5)) # validation
            test_pos = list(set(remanining_pos) - set(val_pos)) # test
        else:
            train_pos = []; val_pos = []; test_pos = []
        if len(negative_cells) > 0:
            train_neg = random.sample(negative_cells, round(len(negative_cells) * 0.6))
            remanining_neg = list(set(negative_cells) - set(train_neg))
            val_neg = random.sample(remanining_neg, round(len(remanining_neg) * 0.5))
            test_neg = list(set(remanining_neg) - set(val_neg))
        else:
            train_neg = []; val_neg = []; test_neg = []


        # loop through training dataset and save images and masks
        newname_train = list(range(len(train_pos) + len(train_neg))); random.shuffle(newname_train)
        train_pos_name = newname_train[:len(train_pos)]; train_neg_name = newname_train[len(train_pos):]

        if len (train_pos_name) > 0:
            for i, j in zip( train_pos_name, train_pos):
                m, im = pos_filter (j)
                # save image
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'training' / f"{i}_img.tif"
                tifffile.imwrite(fPath,im)
                # associated mask
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'training' / f"{i}_mask.tif"
                tifffile.imwrite(fPath, m)

        if len (train_neg_name) > 0:
            for k, l in zip( train_neg_name, train_neg):
                m, im = neg_filter (l)
                # save image
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'training' / f"{k}_img.tif"
                tifffile.imwrite(fPath, im)
                # associated mask
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'training' / f"{k}_mask.tif"
                tifffile.imwrite(fPath, m)


        # loop through validation dataset and save images and masks
        newname_train = list(range(len(val_pos) + len(val_neg))); random.shuffle(newname_train)
        train_pos_name = newname_train[:len(val_pos)]; train_neg_name = newname_train[len(val_pos):]

        if len (train_pos_name) > 0:
            for i, j in zip( train_pos_name, val_pos):
                m, im = pos_filter (j)
                # save image
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'validation' / f"{i}_img.tif"
                tifffile.imwrite(fPath, im)
                # associated mask
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'validation' / f"{i}_mask.tif"
                tifffile.imwrite(fPath, m)

        if len (train_neg_name) > 0:
            for k, l in zip( train_neg_name, val_neg):
                m, im = neg_filter (l)
                # save image
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'validation' / f"{k}_img.tif"
                tifffile.imwrite(fPath, im)
                # associated mask
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'validation' / f"{k}_mask.tif"
                tifffile.imwrite(fPath, m)


        # loop through test dataset and save images and masks
        newname_train = list(range(len(test_pos) + len(test_neg))); random.shuffle(newname_train)
        train_pos_name = newname_train[:len(test_pos)]; train_neg_name = newname_train[len(test_pos):]

        if len (train_pos_name) > 0:
            for i, j in zip( train_pos_name, test_pos):
                m, im = pos_filter (j)
                # save image
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'test' / f"{i}_img.tif"
                tifffile.imwrite(fPath, im)
                # associated mask
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'test' / f"{i}_mask.tif"
                tifffile.imwrite(fPath, m)

        if len (train_neg_name) > 0:
            for k, l in zip( train_neg_name, test_neg):
                m, im = neg_filter (l)
                # save image
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'test' / f"{k}_img.tif"
                tifffile.imwrite(fPath, im)
                # associated mask
                fPath = projectDir / 'CSPOT/TrainingData/' / f"{marker_name}" / 'test' / f"{k}_mask.tif"
                tifffile.imwrite(fPath, m)

    # apply function to all folders
    r_findFiles = lambda x: findFiles (folderIndex=x)
    process_folders = list(map(r_findFiles, list(range(len(thumbnailFolder)))))

    # Print
    if verbose is True:
        print('Training data has been generated, head over to "' + str(projectDir) + '/CSPOT/TrainingData" to view results')


# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate train, test, and validation cohorts from human sorted thumbnails.')
    parser.add_argument('--thumbnailFolder', type=str, nargs='+', help='List of folders that contains the human sorted Thumbnails that is to be used for generating training data and split them train test and validation cohorts.')
    parser.add_argument('--projectDir', type=str, help='Path to output directory.')
    parser.add_argument('--file_extension', type=str, default=None, help='If there are non-image files in the thumbnailFolder, the user can specify a file extension to only select those files for processing.')
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")    
    parser.add_argument('--TruePos', type=str, default='TruePos', help='Name of the folder that holds the Thumbnails classified as True Positive.')
    parser.add_argument('--NegToPos', type=str, default='NegToPos', help='Name of the folder that holds the Thumbnails classified as True Negative.')
    parser.add_argument('--TrueNeg', type=str, default='TrueNeg', help='Name of the folder that holds the Thumbnails that were moved from `True Positive` to `True Negative`.')
    parser.add_argument('--PosToNeg', type=str, default='PosToNeg', help='Name of the folder that holds the Thumbnails that were moved from `True Negative` to `True Positive`.')
    args = parser.parse_args()
    generateTrainTestSplit(thumbnailFolder=args.thumbnailFolder,
                           projectDir=args.projectDir,
                           file_extension=args.file_extension,
                           verbose=args.verbose,
                           TruePos=args.TruePos,
                           NegToPos=args.NegToPos, 
                           TrueNeg=args.TrueNeg,
                           PosToNeg=args.PosToNeg)




# Libs
import tifffile
import numpy as np
from skimage import io as skio

# Functions
def tifread(path):
    return tifffile.imread(path)

def normalize(I):
    m = np.min(I)
    M = np.max(I)
    if M > m:
        return (I-m)/(M-m)
    else:
        return I
    
def im2double(I):
    if I.dtype == 'uint16':
        return I.astype('float64')/65535
    elif I.dtype == 'uint8':
        return I.astype('float64')/255
    elif I.dtype == 'float32':
        return I.astype('float64')
    elif I.dtype == 'float64':
        return I
    else:
        print('returned original image type: ', I.dtype)
        return I
      
def imwrite(I,path):
    skio.imsave(path,I)




"""
!!! abstract "Short Description"
    The function trains a deep learning model for each marker in the provided 
    training data. To train the `cspotModel`, simply direct the function to the 
    `TrainingData` folder. To train only specific models, specify the folder names 
    using the `trainMarkers` parameter. The `projectDir` remains constant and the 
    program will automatically create subfolders to save the trained models.
  

## Function
"""


# Libs
import argparse
import pathlib
if __name__ == '__main__':
    from UNet import *
else:
    from .UNet import UNet2D

# Function
def csTrain(trainingDataPath,
               projectDir,
               trainMarkers=None,
               artefactPath=None,
               imSize=64,
               nChannels=1,
               nClasses=2,
               nExtraConvs=0,
               nLayers=3,
               featMapsFact=2,
               downSampFact=2,
               ks=3,
               nOut0=16,
               stdDev0=0.03,
               batchSize=16,
               epochs=100,
               verbose=True):
    """

Parameters:
    trainingDataPath (str):
        The file path leading to the directory that holds the training data.
    
    projectDir (str):
        Path to output directory. The result will be located at `projectDir/CSPOT/cspotModel/`.
    
    trainMarkers (list):
        Generate models for a specified list of markers. By default, models are c
        reated for all data in the TrainingData folder. If the user wants to
        limit it to a specific list, they can pass in the folder names (e.g. ['CD3D', 'CD4'])
    
    artefactPath (str):
        Path to the directory where the artefacts data is loaded from.
    
    imSize (int, optional):
        Image size (assumed to be square).
    
    nChannels (int, optional):
        Number of channels in the input image.
    
    nClasses (int, optional):
        Number of classes in the classification problem.
    
    nExtraConvs (int, optional):
        Number of extra convolutional layers to add to the model.
    
    nLayers (int, optional):
        Total number of layers in the model.
    
    featMapsFact (int, optional):
        Factor to multiply the number of feature maps by in each layer.
    
    downSampFact (int, optional):
        Factor to down-sample the feature maps by in each layer.
    
    ks (int, optional):
        Kernel size for the convolutional layers.
    
    nOut0 (int, optional):
        Number of filters in the first layer.
    
    stdDev0 (float, optional):
        Standard deviation for the initializer for the first layer.
    
    batchSize (int, optional):
        Batch size for training.
    
    epochs (int, optional):
        Number of training epochs.

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  

Returns:
    Model (images and model):  
        The result will be located at `projectDir/CSPOT/cspotModel/`.
    

Example:
    ```python
    
    # High level working directory
    projectDir = '/Users/aj/Documents/cspotExampleData'
    
    trainingDataPath = projectDir + '/CSPOT/TrainingData'
    
    cs.csTrain(trainingDataPath=trainingDataPath,
                   projectDir=projectDir,
                   trainMarkers=None,
                   artefactPath=None,
                   imSize=64,
                   nChannels=1,
                   nClasses=2,
                   nExtraConvs=0,
                   nLayers=3,
                   featMapsFact=2,
                   downSampFact=2,
                   ks=3,
                   nOut0=16,
                   stdDev0=0.03,
                   batchSize=16,
                   epochs=1)
    
    # Same function if the user wants to run it via Command Line Interface
    python csTrain.py \
        --trainingDataPath /Users/aj/Documents/cspotExampleData/CSPOT/TrainingData \
        --projectDir /Users/aj/Documents/cspotExampleData/ \
        --epochs 1
    
    ```


    """

    # Start here
    # convert to path
    trainingDataPath = pathlib.Path(trainingDataPath)
    # identify all the data folders within the given TrainingData folder
    directories = [x for x in trainingDataPath.iterdir() if x.is_dir()]
    # keep only folders that the user have requested
    if trainMarkers is not None:
        if isinstance(trainMarkers, str):
            trainMarkers = [trainMarkers]
        directories = [x for x in directories if x.stem in trainMarkers]

    # optional artifacts
    if artefactPath is not None:
        artefactPath = pathlib.Path(artefactPath)
        artefactTrainPath = pathlib.Path(artefactPath / 'training')
        artefactValidPath = pathlib.Path(artefactPath / 'validation')
    else:
        artefactPath = ''
        artefactTrainPath = ''
        artefactValidPath = ''
    # Need to run the training for each marker

    def csTrainInternal(trainingDataPath,
                           projectDir,
                           artefactPath,
                           imSize,
                           nChannels,
                           nClasses,
                           nExtraConvs,
                           nLayers,
                           featMapsFact,
                           downSampFact,
                           ks,
                           nOut0,
                           stdDev0,
                           batchSize,
                           epochs):
        # process the file name
        finalName = trainingDataPath.stem
        
        # paths for loading data
        trainPath = pathlib.Path(trainingDataPath / 'training')
        validPath = pathlib.Path(trainingDataPath / 'validation')
        testPath = pathlib.Path(trainingDataPath / 'test')

        # Paths for saving data
        logPath = pathlib.Path(
            projectDir + '/CSPOT/csTrain/' + finalName + '/tempTFLogs/')
        modelPath = pathlib.Path(projectDir + '/CSPOT/cspotModel/' + finalName)
        pmPath = pathlib.Path(projectDir + '/CSPOT/csTrain/' +
                              finalName + '/TFprobMaps/')

        # set up the model
        UNet2D.setup(imSize=imSize,
                     nClasses=nClasses,
                     nChannels=nChannels,
                     nExtraConvs=nExtraConvs,
                     nDownSampLayers=nLayers,
                     featMapsFact=featMapsFact,
                     downSampFact=downSampFact,
                     kernelSize=ks,
                     nOut0=nOut0,
                     stdDev0=stdDev0,
                     batchSize=batchSize)

        # train the model
        UNet2D.train(trainPath=trainPath,
                     validPath=validPath,
                     testPath=testPath,
                     artTrainPath=artefactTrainPath,
                     artValidPath=artefactValidPath,
                     logPath=logPath,
                     modelPath=modelPath,
                     pmPath=pmPath,
                     restoreVariables=False,
                     nSteps=epochs,
                     gpuIndex=0,
                     testPMIndex=2)

    # Run the function on all markers
    def r_csTrainInternal(x): return csTrainInternal(trainingDataPath=x,
                                                           projectDir=projectDir,
                                                           artefactPath=artefactPath,
                                                           imSize=imSize,
                                                           nChannels=nChannels,
                                                           nClasses=nClasses,
                                                           nExtraConvs=nExtraConvs,
                                                           nLayers=nLayers,
                                                           featMapsFact=featMapsFact,
                                                           downSampFact=downSampFact,
                                                           ks=ks,
                                                           nOut0=nOut0,
                                                           stdDev0=stdDev0,
                                                           batchSize=batchSize,
                                                           epochs=epochs)

    csTrainInternal_result = list(
        map(r_csTrainInternal,  directories))  # Apply function
    
    # Finish Job
    if verbose is True:
        print('CSPOT Models have been generated, head over to "' + str(projectDir) + '/CSPOT/cspotModel" to view results')



if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Train a deep learning model using data for each marker, stored in separate directories.')
    parser.add_argument('--trainingDataPath', type=str, help='The file path leading to the directory that holds the training data for each marker.')
    parser.add_argument('--projectDir', type=str, help='Path to output directory.')
    parser.add_argument("--trainMarkers", type=str, nargs="*", default=None, help="Generate models for a specified list of markers.")
    parser.add_argument('--artefactPath', type=str, help='Path to the directory where the artefacts data is loaded from.', default=None)
    parser.add_argument('--imSize', type=int, help='Image size (assumed to be square).', default=64)
    parser.add_argument('--nChannels', type=int, help='Number of channels in the input image.', default=1)
    parser.add_argument('--nClasses', type=int, help='Number of classes in the classification problem.', default=2)
    parser.add_argument('--nExtraConvs', type=int, help='Number of extra convolutional layers to add to the model.', default=0)
    parser.add_argument('--nLayers', type=int, help='Total number of layers in the model.', default=3)
    parser.add_argument('--featMapsFact', type=int, help='Factor to multiply the number of feature maps by in each layer.', default=2)
    parser.add_argument('--downSampFact', type=int, help='Factor to down-sample the feature maps by in each layer.', default=2)
    parser.add_argument('--ks', type=int, help='Kernel size for the convolutional layers.', default=3)
    parser.add_argument('--nOut0', type=int, help='Number of filters in the first layer.', default=16)
    parser.add_argument('--stdDev0', type=float, help='Standard deviation for the initializer for the first layer.', default=0.03)
    parser.add_argument('--batchSize', type=int, help='Batch size for training.', default=16)
    parser.add_argument('--epochs', type=int, help='Number of training epochs.', default=100)
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")    

    args = parser.parse_args()
    csTrain (trainingDataPath=args.trainingDataPath,
               projectDir=args.projectDir,
               trainMarkers=args.trainMarkers,
               artefactPath=args.artefactPath,
               imSize=args.imSize,
               nChannels=args.nChannels,
               nClasses=args.nClasses,
               nExtraConvs=args.nExtraConvs,
               nLayers=args.nLayers,
               featMapsFact=args.featMapsFact,
               downSampFact=args.downSampFact,
               ks=args.ks,
               nOut0=args.nOut0,
               stdDev0=args.stdDev0,
               batchSize=args.batchSize,
               epochs=args.epochs,
               verbose=args.verbose)



import numpy as np
from os.path import exists, isfile, join, split, splitext, isdir
from os import makedirs, listdir
import tifffile
import matplotlib.pyplot as plt
from skimage import io as skio
from skimage import transform as trfm
from skimage.morphology import binary_erosion, disk
import shutil
import pandas as pd
import pickle
from torchvision.transforms import functional as F
import random
import torch.distributed as dist
from scipy.ndimage import gaussian_filter
from scipy.ndimage.morphology import binary_fill_holes

class Compose(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, image, target):
        for t in self.transforms:
            image, target = t(image, target)
        return image, target

class RandomHorizontalFlip(object):
    def __init__(self, prob):
        self.prob = prob

    def __call__(self, image, target):
        if random.random() < self.prob:
            height, width = image.shape[-2:]
            image = image.flip(-1)
            bbox = target["boxes"]
            bbox[:, [0, 2]] = width - bbox[:, [2, 0]]
            target["boxes"] = bbox
            if "masks" in target:
                target["masks"] = target["masks"].flip(-1)
            if "keypoints" in target:
                keypoints = target["keypoints"]
                keypoints = _flip_coco_person_keypoints(keypoints, width)
                target["keypoints"] = keypoints
        return image, target

class ToTensor(object):
    def __call__(self, image, target):
        image = F.to_tensor(image)
        return image, target

def get_transform(train):
    transforms = []
    # converts the image, a PIL image, into a PyTorch Tensor
    transforms.append(ToTensor())
    if train:
        # during training, randomly flip the training images
        # and ground-truth for data augmentation
        transforms.append(RandomHorizontalFlip(0.5))
    return Compose(transforms)

def collate_fn(batch):
    return tuple(zip(*batch))

def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True

def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()

def reduce_dict(input_dict, average=True):
    """
    Args:
        input_dict (dict): all the values will be reduced
        average (bool): whether to do average or sum
    Reduce the values in the dictionary from all processes so that all processes
    have the averaged results. Returns a dict with the same fields as
    input_dict, after reduction.
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict
    with torch.no_grad():
        names = []
        values = []
        # sort the keys so that they are consistent across processes
        for k in sorted(input_dict.keys()):
            names.append(k)
            values.append(input_dict[k])
        values = torch.stack(values, dim=0)
        dist.all_reduce(values)
        if average:
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}
    return reduced_dict

def createFolderIfNonExistent(path):
    if not exists(path): # from os.path
        makedirs(path)
        
def removeFolderIfExistent(path):
    if exists(path):
        shutil.rmtree(path)

def tifread(path):
    return tifffile.imread(path)

def tifwrite(I,path):
    tifffile.imsave(path, I)

def imshow(I,**kwargs):
    if not kwargs:
        plt.imshow(I,cmap='gray')
    else:
        plt.imshow(I,**kwargs)
        
    plt.axis('off')
    plt.show()

def imread(path):
    return skio.imread(path)

def imwrite(I,path):
    return skio.imsave(path,I)

def imerode(I,r):
    return binary_erosion(I, disk(r))

def im2double(I):
    if I.dtype == 'uint16':
        return I.astype('float64')/65535
    elif I.dtype == 'uint8':
        return I.astype('float64')/255
    elif I.dtype == 'float32':
        return I.astype('float64')
    elif I.dtype == 'float64':
        return I
    else:
        print('returned original image type: ', I.dtype)
        return I

def imresizeDouble(I,sizeOut): # input and output are double
    return trfm.resize(I,(sizeOut[0],sizeOut[1]),mode='reflect')
    
def listfiles(path,token): # path = folder path
    l = []
    for f in listdir(path):
        fullPath = join(path,f)
        if isfile(fullPath) and token in f:
            l.append(fullPath)
    l.sort()
    return l

def fileparts(path): # path = file path
    [p,f] = split(path)
    [n,e] = splitext(f)
    return [p,n,e]

def pathjoin(p,ne): # '/path/to/folder', 'name.extension' (or a subfolder)
    return join(p,ne)

def imadjust(I):
    I0 = I[I > 0]
    if np.any(I0):
        p1 = np.percentile(I0,1)
        p99 = np.percentile(I0,99)
    else:
        p1 = np.min(I)
        p99 = np.max(I)
        if p99 == 0:
            p99 = 1
    I = (I-p1)/(p99-p1)
    I[I < 0] = 0
    I[I > 1] = 1
    return I

def uint16Gray_to_uint8RGB(I):
    assert I.dtype == 'uint8'
    I = imadjust(I.astype('float64')/255)
    return np.uint8(255*np.stack([I,I,I],axis=2))

def uint16Gray_to_doubleGray(I):
    assert I.dtype == 'uint16'
    return imadjust(I.astype('float64')/65535)

def doubleGray_to_uint8RGB(I):
    assert I.dtype == 'float64'
    return np.uint8(255*np.stack([I,I,I],axis=2))

def saveData(data,path,verbose=False):
    if verbose:
        print('saving data')
    dataFile = open(path, 'wb')
    pickle.dump(data, dataFile)

def loadData(path,verbose=False):
    if verbose:
        print('loading data')
    dataFile = open(path, 'rb')
    return pickle.load(dataFile)

def writeTable(path,colTitles,matrix):
    # with open(path, 'w') as csvfile:
    #     writer = csv.writer(csvfile)
    #     writer.writerow(colTitles)
    #     for i in range(matrix.shape[0]):
    #         writer.writerow(matrix[i,:])

    T = {}
    for i in range(len(colTitles)):
        T[colTitles[i]] = matrix[:,i]
    df = pd.DataFrame(T)
    df.to_csv(path, index=False)

def readTable(path):
    # T = []
    # with open(path) as csvfile:
    #     reader = csv.reader(csvfile)
    #     for row in reader:
    #         T.append(row)
    # colTitles = T[0]
    # nRows = len(T)-1
    # nCols = len(colTitles)
    # matrix = np.zeros((nRows,nCols))
    # for i in range(nRows):
    #     for j in range(nCols):
    #         matrix[i,j] = T[i+1][j]
    # return colTitles, matrix

    df = pd.read_csv(path)
    colTitles = df.columns.to_list()
    matrix = df.to_numpy()
    return colTitles, matrix

def imgaussfilt(I,sigma,**kwargs):
    return gaussian_filter(I,sigma,**kwargs)

def boxes_intersect(box_a, box_b):
    xmin_a, ymin_a, xmax_a, ymax_a = box_a
    xmin_b, ymin_b, xmax_b, ymax_b = box_b

    min_ymax = np.minimum(ymax_a, ymax_b)
    max_ymin = np.maximum(ymin_a, ymin_b)
    min_xmax = np.minimum(xmax_a, xmax_b)
    max_xmin = np.maximum(xmin_a, xmin_b)

    x_intersection = np.maximum(min_xmax-max_xmin, 0)
    y_intersection = np.maximum(min_ymax-max_ymin, 0)
    
    return x_intersection*y_intersection > 0

def imfillholes(I):
    return binary_fill_holes(I)


# -*- coding: utf-8 -*-
#Created on Wed Aug 31 15:08:38 2022
#@author: Ajit Johnson Nirmal
#When LocalNorm is used to curate the thumbnails, use this function to clone the curation on the real files

"""
!!! abstract "Short Description"
    The purpose of the `cloneFolder` function is to copy user actions from one 
    folder to another. For example, if a user manually arranges thumbnails in 
    the `localNorm` folder, this function can replicate those changes to the 
    raw thumbnails.


## Function
"""

# lib
import pathlib
from os import walk
import os
import shutil
import argparse

# function
def cloneFolder (copyFolder, 
                 applyFolder,
                 TruePos='TruePos', 
                 TrueNeg='TrueNeg',
                 PosToNeg='PosToNeg', 
                 NegToPos='NegToPos',
                 verbose=True):
    """
Parameters:
    copyFolder (list):
        List of folders from which the user wants to replicate the file structure.

    applyFolder (list):
        List of folders where the replicated file structure should be applied,
        in the same order as the `copyFolder` list.

    TruePos (str, optional):
        Name of the folder that holds the Thumbnails classified as True Positive.

    TrueNeg (str, optional):
        Name of the folder that holds the Thumbnails classified as True Negative.

    PosToNeg (str, optional):
        Name of the folder that holds the Thumbnails that were moved from `True Positive`
        to `True Negative`.

    NegToPos (str, optional):
        Name of the folder that holds the Thumbnails that were moved from `True Negative`
        to `True Positive`.

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  

Returns:
    folder (cloned folders):  
        The file structure of the source Folder is replicated in the destination Folder.

Example:
        ```python
        
        # High level working directory
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        # list of folders to copy settings from
        copyFolder = [projectDir + '/CSPOT/Thumbnails/localNorm/CD3D',
                      projectDir + '/CSPOT/Thumbnails/localNorm/ECAD']
        # list of folders to apply setting to
        applyFolder = [projectDir + '/CSPOT/Thumbnails/CD3D',
                       projectDir + '/CSPOT/Thumbnails/ECAD']
        # note: Every copyFolder should have a corresponding applyFolder. The order matters! 
        
        # The function accepts the four pre-defined folders. If you had renamed them, please change it using the parameter below.
        cs.cloneFolder (copyFolder, 
                        applyFolder, 
                        TruePos='TruePos', TrueNeg='TrueNeg', 
                        PosToNeg='PosToNeg', NegToPos='NegToPos')
        
                
        # Same function if the user wants to run it via Command Line Interface
        python cloneFolder.py \
            --copyFolder /Users/aj/Desktop/cspotExampleData/CSPOT/Thumbnails/localNorm/CD3D /Users/aj/Desktop/cspotExampleData/CSPOT/Thumbnails/localNorm/ECAD \
            --applyFolder /Users/aj/Desktop/cspotExampleData/CSPOT/Thumbnails/CD3D /Users/aj/Desktop/cspotExampleData/CSPOT/Thumbnails/ECAD
        
        ```

    """

    #TruePos='TruePos'; TrueNeg='TrueNeg'; PosToNeg='PosToNeg'; NegToPos='NegToPos'

    # Convert the path to list
    if isinstance (copyFolder, str):
        copyFolder = [copyFolder]
    if isinstance (applyFolder, str):
        applyFolder = [applyFolder]

    # Quick Check!
    if len(copyFolder) is not len(applyFolder):
        raise ValueError('The number of copyFolder and applyFolder should match, please check!' )

    # function to delete images
    def deleteFile(files, location):
        for f in files:
            # full path
            #full_path = location  + f
            full_path = pathlib.Path.joinpath(location, f)
            if os.path.exists(full_path):
                os.remove(full_path)

    # Function to move images
    def moveFile(files, from_loc, to_loc):
        for f in files:
            # full path
            full_path_from = pathlib.Path.joinpath(from_loc, f) # from_loc + f
            full_path_to = pathlib.Path.joinpath(to_loc, f) # to_loc + f
            # move file
            if os.path.exists(full_path_from):
                shutil.move(full_path_from, full_path_to)

    # path lib of all folder
    all_folders = [pathlib.Path(p) for p in copyFolder]

    # copy from location
    pos_aug_location = [pathlib.Path(p + '/' + str(TruePos)) for p in copyFolder]
    neg_aug_location = [pathlib.Path(p + '/' + str(TrueNeg)) for p in copyFolder]
    pos2neg_aug_location = [pathlib.Path(p + '/' + str(PosToNeg)) for p in copyFolder]
    neg2pos_aug_location = [pathlib.Path(p + '/' + str(NegToPos)) for p in copyFolder]

    # copy to location
    pos_real_location = [pathlib.Path(p + '/' + str(TruePos)) for p in applyFolder]
    neg_real_location = [pathlib.Path(p + '/' + str(TrueNeg)) for p in applyFolder]
    pos2neg_real_location = [pathlib.Path(p + '/' + str(PosToNeg)) for p in applyFolder]
    neg2pos_real_location = [pathlib.Path(p + '/' + str(NegToPos)) for p in applyFolder]



    # function
    def processFolder (folderIndex):
        if verbose is True:
            print ('Processing: ' + str(all_folders[folderIndex].stem))

        # create a list of all file names in the applyFolder
        pos_files = next(walk(pos_real_location[folderIndex]), (None, None, []))[2]
        neg_files = next(walk(neg_real_location[folderIndex]), (None, None, []))[2]

        # Find file names within each of the copyFolder
        pos = next(walk(pos_aug_location[folderIndex]), (None, None, []))[2]
        neg = next(walk(neg_aug_location[folderIndex]), (None, None, []))[2]
        pos2neg = next(walk(pos2neg_aug_location[folderIndex]), (None, None, []))[2]
        neg2pos = next(walk(neg2pos_aug_location[folderIndex]), (None, None, []))[2]  # [] if no file

        # Find images to delete
        pos_del = list(set(pos_files).difference(pos + pos2neg))
        neg_del = list(set(neg_files).difference(neg + neg2pos))

        # delete files
        deleteFile(files=pos_del, location=pos_real_location[folderIndex])
        deleteFile(files=neg_del, location=neg_real_location[folderIndex])

        # move files
        moveFile (files=pos2neg, from_loc=pos_real_location[folderIndex], to_loc=pos2neg_real_location[folderIndex])
        moveFile (files=neg2pos, from_loc=neg_real_location[folderIndex], to_loc=neg2pos_real_location[folderIndex])

        # print the number of files
        posaug = len(next(walk(pos_aug_location[folderIndex]), (None, None, []))[2])
        posreal = len(next(walk(pos_real_location[folderIndex]), (None, None, []))[2])
        negaug = len(next(walk(neg_aug_location[folderIndex]), (None, None, []))[2])
        negreal = len(next(walk(neg_real_location[folderIndex]), (None, None, []))[2])
        postonegaug = len(next(walk(pos2neg_aug_location[folderIndex]), (None, None, []))[2])
        postonegreal = len(next(walk(pos2neg_real_location[folderIndex]), (None, None, []))[2])
        negtoposaug = len(next(walk(neg2pos_aug_location[folderIndex]), (None, None, []))[2])
        negtoposreal = len(next(walk(neg2pos_real_location[folderIndex]), (None, None, []))[2])

        #print ('No of Files in TruePos-> copyFolder: ' + str(posaug) + ' ; applyFolder: '+ str(posreal))
        #print ('No of Files in TrueNeg-> copyFolder: ' + str(negaug) + ' ; applyFolder: '+ str(negreal))
        #print ('No of Files in PosToNeg-> copyFolder: ' + str(postonegaug) + ' ; applyFolder: '+ str(postonegreal))
        #print ('No of Files in NegToPos-> copyFolder: ' + str(negtoposaug) + ' ; applyFolder: '+ str(negtoposreal))

    # apply function to all folders
    r_processFolder = lambda x: processFolder (folderIndex=x)
    process_folders = list(map(r_processFolder, list(range(len(copyFolder)))))

    # Finish Job
    if verbose is True:
        print('Cloning Folder is complete, head over to /CSPOT/Thumbnails" to view results')

# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Clone folder file structure')
    parser.add_argument('--copyFolder', metavar='copyFolder', type=str, nargs='+', help='List of folders from which to replicate the file structure')
    parser.add_argument('--applyFolder', metavar='applyFolder', type=str, nargs='+', help='List of folders where the replicated file structure should be applied, in the same order as the copyFolder list')
    parser.add_argument('--TruePos', dest='TruePos', type=str, default='TruePos', help='Name of the folder that holds the Thumbnails classified as True Positive')
    parser.add_argument('--TrueNeg', dest='TrueNeg', type=str, default='TrueNeg', help='Name of the folder that holds the Thumbnails classified as True Negative')
    parser.add_argument('--PosToNeg', dest='PosToNeg', type=str, default='PosToNeg', help='Name of the folder that holds the Thumbnails that were moved from True Positive to True Negative')
    parser.add_argument('--NegToPos', dest='NegToPos', type=str, default='NegToPos', help='Name of the folder that holds the Thumbnails that were moved from True Negative to True Positive')
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")   
    args = parser.parse_args()
    cloneFolder(copyFolder=args.copyFolder,
                applyFolder=args.applyFolder,
                TruePos=args.TruePos,
                TrueNeg=args.TrueNeg,
                PosToNeg=args.PosToNeg,
                NegToPos=args.NegToPos,
                verbose=args.verbose)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 17 11:56:32 2022
@author: Ajit Johnson Nirmal
Gator Algorithm
"""

import pandas as pd
import anndata as ad
from sklearn.preprocessing import RobustScaler
from sklearn.ensemble import HistGradientBoostingClassifier
from sklearn.neighbors import BallTree
import random
import numpy as np
from sklearn.mixture import GaussianMixture
from collections import defaultdict
from sklearn.neighbors import LocalOutlierFactor
from sklearn.preprocessing import StandardScaler
import pathlib
import os


def gator (gatorObject,
           probQuant='probQuant',
           minAbundance=0.002,
           percentiles=[1, 20, 80, 99],
           drop_markers = None, 
           leakData=True,
           scaleData=True,
           log=True,
           x_coordinate='X_centroid',
           y_coordinate='Y_centroid',
           imageid='imageid',
           seed=0,
           rescaleMethod='minmax',
           label='gatorPredict', 
           silent=True,
           outputDir=None, **kwargs):
    
    #gatorObject = "/Users/aj/Dropbox (Partners HealthCare)/Data/gator/data/ajn_training_data/GATOR/gatorObject/1_6_GatorOutput.h5ad"
    #probQuant='probQuant'; minAbundance=0.002; percentiles=[1, 20, 80, 99]; drop_markers = None;leakData=False;rescaleMethod='minmax';silent = False
    #scaleData=False; log=True; x_coordinate='X_centroid'; y_coordinate='Y_centroid'; imageid='imageid'; seed=0; label='gatorPredict'
    #gatorObject = '/Users/aj/Dropbox (Partners HealthCare)/Data/gator/data/ajn_training_data/GATOR/gatorPredict/2_28_GatorOutput.h5ad'
    #gatorObject = '/Users/aj/Dropbox (Partners HealthCare)/Data/gator/data/ajn_training_data/GATOR/gatorPredict/4_113_GatorOutput.h5ad'


    # Load the andata object    
    if isinstance(gatorObject, str):
        adata = ad.read(gatorObject)
        gatorObject = [gatorObject]
        gatorObjectPath = [pathlib.Path(p) for p in gatorObject]
    else:
        adata = gatorObject.copy()
    
    # break the function if probQuant is not detectable
    def check_key_exists(dictionary, key):
        try:
            # Check if the key exists in the dictionary
            value = dictionary[key]
        except KeyError:
            # Return an error if the key does not exist
            return "Error: " + str(probQuant) + " does not exist, please check!"
    # Test 
    check_key_exists(dictionary=adata.uns, key=probQuant)  
    
    
    ###########################################################################
    # SOME GENERIC FUNCTIONS
    ###########################################################################
    
    # used in (step 1)
    def get_columns_with_low_values(df, minAbundance):
        columns_to_keep = []
        for column in df.columns:
            num_rows_with_high_values = len(df[df[column] > 0.6])
            if num_rows_with_high_values / len(df) < minAbundance:
                columns_to_keep.append(column)
        return columns_to_keep
    
    # preprocess data (step-4)
    def pre_process (data, log=log):
        # clip outliers
        def clipping (x):
            clip = x.clip(lower =np.percentile(x,0.01), upper=np.percentile(x,99.99)).tolist()
            return clip
        processsed_data = data.apply(clipping)
        if log is True:
            processsed_data = np.log1p(processsed_data)
        return processsed_data
    # preprocess data (step-5)
    def apply_transformation (data):
        # rescale the data
        transformer = RobustScaler().fit(data)
        processsed_data = pd.DataFrame(transformer.transform(data), columns = data.columns, index=data.index)
        return processsed_data
    
    # GMM
    def simpleGMM (data, n_components, means_init, random_state):
        gmm = GaussianMixture(n_components=n_components, means_init=means_init,  random_state=random_state)
        gmm.fit(data)
        # Predict the class labels for each sample
        predictions = gmm.predict(data)
        # Get the mean of each Gaussian component
        means = gmm.means_.flatten()
        # Sort the mean values in ascending order
        sorted_means = np.sort(means)
        # Assign 'pos' to rows with higher mean distribution and 'neg' to rows with lower mean distribution
        labels = np.where(predictions == np.argmax(means), 'pos', 'neg')  

        return labels
    
    # take in two list (ccategorical and numerical) and return mean values
    def array_mean (labels, values):
        # Create a defaultdict with an empty list as the default value
        result = defaultdict(list)
        # Iterate over the labels and values arrays
        for label, value in zip(labels, values):
            # Add the value to the list for the corresponding label
            result[label].append(value)
        # Calculate the mean for each label and store it in the dictionary
        for label, value_list in result.items():
            result[label] = np.mean(value_list)
        return result
    
    # match two arrays and return seperate lists
    def array_match (labels, names):
        # Create a defaultdict with an empty list as the default value
        result = defaultdict(list)
        # Iterate over the labels and names arrays
        for label, name in zip(labels, names):
            # Add the name to the list for the corresponding label
            result[label].append(name)
        return result
    
    # return the mean between two percentiles
    def meanPercentile (values, lowPercentile, highPercentile):
        # Calculate the 1st percentile value
        p1 = np.percentile(values, lowPercentile)      
        # Calculate the 20th percentile value
        p20 = np.percentile(values, highPercentile)      
        # Select the values between the 1st and 20th percentile using numpy.where()
        filtered_values = np.where((values >= p1) & (values <= p20)) 
        # Calculate the mean of the filtered values
        meanVal = np.mean(values[filtered_values])
        return meanVal
    
    # return the mean between two percentiles
    def indexPercentile (processed_data, marker, lowPercentile, highPercentile):
        values = processed_data[marker].values
        # Calculate the 1st percentile value
        p1 = np.percentile(values, lowPercentile)      
        # Calculate the 20th percentile value
        p20 = np.percentile(values, highPercentile)      
        # Select the values between the 1st and 20th percentile using numpy.where()
        filtered_values = np.where((values >= p1) & (values <= p20)) 
        # Calculate the mean of the filtered values
        idx = processed_data[marker].iloc[filtered_values].index
        return idx
         
    # Leaking expression inforamtion into neighbouring cells function (step-4)
    def leakDataF (bonafideCells, processed_data, neighbours, marker, seed):
        cells = processed_data.loc[bonafideCells].reset_index(drop=True)#[marker].values  
        cells_neigh = neighbours[neighbours['cell'].isin(bonafideCells)]['neigh'].values        
        # find original expression profile of both the cell and its neighbour
        cells_neigh = processed_data.loc[cells_neigh].reset_index(drop=True)#[marker].values
        # formula leak of 5-40%
        # x = (a + leak*b) / 1+ leak
        random.seed(seed)
        leak = [(random.randint(5, 40) / 100) for _ in range(len(cells))]
        nLeak = [x + 1 for x in leak] # normalizing the cell size        
        def transform_dataframes(cells, cells_neigh, leak, nLeak):
            # Multiply leak vector to all columns of cells_neigh
            cells_neigh_transformed = cells_neigh.multiply(leak, axis=0)
            # Add cells_neigh_transformed to cells element-wise
            cells_transformed = cells.add(cells_neigh_transformed)
            # Divide each column of cells_transformed by 1+leak
            cells_final = cells_transformed.divide(nLeak, axis=0)            
            return cells_final
        # Run the function
        leakData_result = transform_dataframes (cells, cells_neigh, leak, nLeak)
        return leakData_result
    
    # Used for rescaling data
    # used to find the mid point of GMM mixtures
    def find_midpoint(data, labels):
      # Convert data and labels to NumPy arrays
      data = np.array(data)
      labels = np.array(labels)
      # Get the indices that would sort the data and labels arrays
      sort_indices = np.argsort(data)
      # Sort the data and labels arrays using the sort indices
      sorted_data = data[sort_indices]
      sorted_labels = labels[sort_indices]
      # Find the index where the 'neg' and 'pos' labels meet
      midpoint_index = np.argmax(sorted_labels == 'pos')
      # Return the value at the midpoint index
      return sorted_data[midpoint_index]
    
    # Used for reassigning some of the wrong 'nes' and 'pos' within data given a midpoint
    def modify_negatives_vectorized(data, labels, midpoint):
      # Convert data and labels to NumPy arrays
      data = np.array(data)
      labels = np.array(labels)
      # Get the indices that would sort the data and labels arrays
      sort_indices = np.argsort(data)
      # Sort the data and labels arrays using the sort indices
      sorted_data = data[sort_indices]
      sorted_labels = labels[sort_indices]
      # Find the index where the sorted data is greater than or equal to the midpoint value
      midpoint_index = np.argmax(sorted_data >= midpoint)
      # Find all the elements in the sorted labels array with a value of 'neg' after the midpoint index
      neg_mask = np.logical_and(sorted_labels == 'neg', np.arange(len(sorted_data)) >= midpoint_index)
      # Modify the value of the elements to be equal to the midpoint value
      sorted_data[neg_mask] = midpoint
      # Find all the elements in the sorted labels array with a value of 'pos' before the midpoint index
      pos_mask = np.logical_and(sorted_labels == 'pos', np.arange(len(sorted_data)) < midpoint_index)
      # Modify the value of the elements to be equal to the midpoint value plus 0.1
      sorted_data[pos_mask] = midpoint + 0.1
      # Reorder the data array to the original order
      reordered_data = sorted_data[np.argsort(sort_indices)]
      # Return the modified data
      return reordered_data
  
# =============================================================================
#     def modify_prediction_results(rawData, prediction_results, failedMarkersinData):
#         # Identify the index of the maximum value for each column in rawData
#         max_index = rawData.idxmax()
#         # Iterate through the specified columns of rawData
#         for col in failedMarkersinData:
#             # Get the index of the maximum value
#             max_row = max_index[col]
#             # Modify the corresponding index in prediction_results
#             prediction_results[col].at[max_row] = 'pos'
# =============================================================================
    
    def get_second_highest_values(df, failedMarkersinData):
        # get the second largest value for each column in the list
        second_highest_values = df[failedMarkersinData].max()
        # convert the series to a dictionary
        second_highest_values_dict = second_highest_values.to_dict()
        return second_highest_values_dict
    
    
    # sigmoid scaling to convert the data between 0-1 based on the midpoint
    def sigmoid(x, midpoint):
        return 1 / (1 + np.exp(-(x - midpoint)))
    
    # rescale based on min-max neg -> 0-4.9 and pos -> 0.5-1
    def scale_data(data, midpoint):
        below_midpoint = data[data <= midpoint]
        above_midpoint = data[data > midpoint]
        indices_below = np.where(data <= midpoint)[0]
        indices_above = np.where(data > midpoint)[0]
        
        # Scale the group below the midpoint
        min_below = below_midpoint.min()
        max_below = below_midpoint.max()
        range_below = max_below - min_below
        below_midpoint = (below_midpoint - min_below) / range_below
        
        # Scale the group above the midpoint
        if len(above_midpoint) > 0:
            min_above = above_midpoint.min()
            max_above = above_midpoint.max()
            range_above = max_above - min_above
            above_midpoint = (above_midpoint - min_above) / range_above
        else:
            above_midpoint = []
        
        # Re-assemble the data in the original order by using the indices of the values in each group
        result = np.empty(len(data))
        result[indices_below] = below_midpoint * 0.499999999
        if len(above_midpoint) > 0:
            result[indices_above] = above_midpoint * 0.50 + 0.50
        return result

    
    ###########################################################################
    # step-1 : Identify markers that have failed in this dataset
    ###########################################################################

    failed_markers = get_columns_with_low_values (df=adata.uns[probQuant],minAbundance=minAbundance)
    
    if silent is False:
        print('Failed Markers are: ' + ", ".join(str(x) for x in failed_markers))

    ###########################################################################
    # step-2 : Prepare DATA
    ###########################################################################
    
    rawData = pd.DataFrame(adata.raw.X, columns= adata.var.index, index = adata.obs.index)
    rawprocessed = pre_process (rawData, log=log)
    # drop user defined markers; note if a marker is dropped it will not be part of the 
    # final prediction too. Markers that failed although removed from prediction will
    # still be included in the final predicted output as all negative. 
    if drop_markers is not None: 
        if isinstance(drop_markers, str):
            drop_markers = [drop_markers]
        pre_processed_data = rawprocessed.drop(columns=drop_markers)
    else:
        pre_processed_data = rawprocessed.copy()
    
    # also drop failed markers
    failedMarkersinData = list(set(pre_processed_data.columns).intersection(failed_markers))
    
    # final dataset that will be used for prediction
    pre_processed_data = pre_processed_data.drop(columns=failedMarkersinData)
    
    # isolate the unet probabilities
    probQuant_data = adata.uns[probQuant]
    
    # list of markers to process: (combined should match data)
    expression_unet_common = list(set(pre_processed_data.columns).intersection(set(probQuant_data.columns)))
    only_expression = list(set(pre_processed_data.columns).difference(set(probQuant_data.columns)))


    ###########################################################################
    # step-3 : Build the spatial neighbourhood graph
    ###########################################################################
    
    if leakData is True: 
        spatial_data = adata.obs[[x_coordinate,y_coordinate ]]
        # identify the nearest neighbour
        tree = BallTree(spatial_data, leaf_size= 2)
        ind = tree.query(spatial_data, k=2, return_distance= False)
        neighbours = pd.DataFrame(ind.tolist()) # neighbour DF
        neighbours = neighbours.dropna(how='all')
        # map the cell id for the neighbours
        phenomap = dict(zip(list(range(len(ind))), spatial_data.index)) # Used for mapping
        for i in neighbours.columns:
            neighbours[i] = neighbours[i].dropna().map(phenomap, na_action='ignore')
        neighbours.columns = ['cell', 'neigh']
    else:
        neighbours = None
        
        
    ###########################################################################
    # step-4 : Identify a subset of true positive and negative cells    
    ###########################################################################
    
    # marker = 'CD4'
    def bonafide_cells (marker, 
                        expression_unet_common, only_expression, 
                        pre_processed_data, probQuant_data, seed,
                        percentiles):
        
        
        if marker in expression_unet_common:
            if silent is False:
                print("NN marker: " + str(marker))
            # run GMM on probQuant_data
            X = probQuant_data[marker].values.reshape(-1,1)
            # Fit the GMM model to the data
            labels = simpleGMM (data=X, n_components=2, means_init=None, random_state=seed)
            
            # find the mean of the pos and neg cells in expression data given the labels
            values = pre_processed_data [marker].values
            Pmeans = array_mean (labels, values)
            # Format mean to pass into next GMM
            Pmean = np.array([[ Pmeans.get('neg')], [Pmeans.get('pos')]])
             
            # Now run GMM on the expression data
            Y = pre_processed_data[marker].values.reshape(-1,1)
            labelsE = simpleGMM (data=Y, n_components=2, means_init=Pmean, random_state=seed)
            
            # Match the labels and index names to identify which cells are pos and neg
            expCells = array_match (labels=labels, names=pre_processed_data.index)
            probCells = array_match (labels=labelsE, names=pre_processed_data.index)
            # split it 
            expCellsPos = expCells.get('pos', []) ; expCellsNeg = expCells.get('neg', [])
            probCellsPos = probCells.get('pos', []) ; probCellsNeg = probCells.get('neg', [])
            # find common elements
            pos = list(set(expCellsPos).intersection(set(probCellsPos)))
            neg = list(set(expCellsNeg).intersection(set(probCellsNeg)))
            
            # print no of cells
            if silent is False:
                print("POS cells: {} and NEG cells: {}.".format(len(pos), len(neg)))
            
            # check if the length is less than 20 cells and if so add the marker to only_expression
            if len(pos) < 20 or len(neg) < 20: ## CHECK!
                only_expression.append(marker)
                if silent is False:
                    print ("As the number of POS/NEG cells is low for " + str(marker) + ", GMM will fitted using only expression values.")
            

        if marker in only_expression:
            if silent is False:
                print("Expression marker: " + str(marker))
            # Run GMM only on the expression data
            Z = pre_processed_data[marker].values.reshape(-1,1)
            # if user provides manual percentile, use it to intialize the GMM
            if percentiles is not None:
                percentiles.sort() 
                F = pre_processed_data[marker].values
                # mean of cells within defined threshold
                lowerPercent = meanPercentile (values=F, lowPercentile=percentiles[0], highPercentile=percentiles[1])
                higherPercent = meanPercentile (values=F, lowPercentile=percentiles[2], highPercentile=percentiles[3])
                # Format mean to pass into next GMM
                Pmean = np.array([[lowerPercent], [higherPercent]])
                labelsOE = simpleGMM (data=Z, n_components=2, means_init=Pmean, random_state=seed)
            else:
                labelsOE = simpleGMM (data=Z, n_components=2, means_init=None, random_state=seed)
            # match labels with indexname
            OEcells = array_match (labels=labelsOE, names=pre_processed_data.index)
            # split it
            pos = OEcells.get('pos', []) ; neg = OEcells.get('neg', [])   
            
            # randomly subset 70% of the data to return
            random.seed(seed); pos = random.sample(pos, k=int(len(pos) * 0.7))
            random.seed(seed); neg = random.sample(neg, k=int(len(neg) * 0.7)) 
                
            # print no of cells
            if silent is False:
                print("Defined POS cells is {} and NEG cells is {}.".format(len(pos), len(neg)))
            
            # What happens of POS/NEG is less than 20
            # check if the length is less than 20 cells and if so add the marker to only_expression
            if len(pos) < 20 or len(neg) < 20:  ## CHECK!
                if percentiles is None:
                    percentiles = [1,20,80,99]
                neg = list(indexPercentile (pre_processed_data, marker, lowPercentile=percentiles[0], highPercentile=percentiles[1]))
                pos = list(indexPercentile (pre_processed_data, marker, lowPercentile=percentiles[2], highPercentile=percentiles[3]))
                if silent is False:
                    print ("As the number of POS/NEG cells is low for " + str(marker) + ", cells falling within the given percentile " + str(percentiles) + ' was used.')      

        # return the output
        return marker, pos, neg
    
    # Run the function on all markers
    if silent is False:
        print("Intial GMM Fitting")
    r_bonafide_cells = lambda x: bonafide_cells (marker=x, 
                                                expression_unet_common=expression_unet_common, 
                                                only_expression=only_expression, 
                                                pre_processed_data=pre_processed_data, 
                                                probQuant_data=probQuant_data, 
                                                seed=seed,
                                                percentiles=percentiles)
    bonafide_cells_result = list(map(r_bonafide_cells,  pre_processed_data.columns)) # Apply function  


    ###########################################################################
    # step-5 : Generate training data for the Gradient Boost Classifier
    ###########################################################################
    
    # bonafide_cells_result = bonafide_cells_result[0]
    def trainingData (bonafide_cells_result, pre_processed_data, neighbours, leakData, scaleData):
        # uravel the data
        marker = bonafide_cells_result[0]
        pos = bonafide_cells_result[1]
        neg = bonafide_cells_result[2]
        PD = pre_processed_data.copy()
        
        if silent is False:
            print('Processing: ' + str(marker))
        
        # class balance the number of pos and neg cells based on the lowest denominator
        if len(neg) < len(pos):
            pos = random.sample(pos, len(neg))
        else:
            neg = random.sample(neg, len(pos))
             
        # processed data with pos and neg info
        
        
        #PD['label'] = ['pos' if index in pos else 'neg' if index in neg else 'other' for index in PD.index]
        PD['label'] = np.where(PD.index.isin(pos), 'pos', np.where(PD.index.isin(neg), 'neg', 'other'))


        
        if leakData is True:
            # Now based on the bonafide cells, leak the neighbour information into these cells
            # Leak information into negative cells
            negLeakedDF = leakDataF (bonafideCells=neg, 
                                   processed_data=pre_processed_data, 
                                   neighbours=neighbours,
                                   marker=marker,
                                   seed=seed)
            new_index_names = ['nnu_{}'.format(idx) for idx in negLeakedDF.index]
            negLeakedDF.index = new_index_names
            negLeakedDF['label'] = 'neg'
            
            # Leak information into positive cells
            posLeakedDF = leakDataF (bonafideCells=pos, 
                                   processed_data=pre_processed_data, 
                                   neighbours=neighbours,
                                   marker=marker,
                                   seed=seed)   
            new_index_names = ['npu_{}'.format(idx) for idx in posLeakedDF.index]
            posLeakedDF.index = new_index_names
            posLeakedDF['label'] = 'pos'
            
            # combine all cells togeather
            combined_data = pd.concat([PD, negLeakedDF, posLeakedDF])
        else:
            combined_data = PD
        
        # scale data if requested
        if scaleData is True:
            combined_data_labels = combined_data[['label']]
            combined_data = combined_data.drop('label', axis=1)
            combined_data = apply_transformation(combined_data)
            combined_data = pd.concat ([combined_data, combined_data_labels], axis=1)
        
        # return final output
        return marker, combined_data
    
    # Run the function
    if silent is False:
        print("Building the Training Data")
    r_trainingData = lambda x: trainingData (bonafide_cells_result=x, 
                                                 pre_processed_data=pre_processed_data, 
                                                 neighbours=neighbours, 
                                                 leakData = leakData,
                                                 scaleData=scaleData)
    trainingData_result = list(map(r_trainingData, bonafide_cells_result)) # Apply function  


    ###########################################################################
    # step-6 : Train and Predict on all cells
    ###########################################################################
    
    # trainingData_result = trainingData_result[2]
    def gatorClassifier (trainingData_result):
        
        #unravel data
        marker = trainingData_result[0]
        combined_data = trainingData_result[1]
        
        # prepare the data for predicition
        index_names_to_drop = [index for index in combined_data.index if 'npu' in index or 'nnu' in index]
        predictionData = combined_data.drop(index=index_names_to_drop, inplace=False)
        predictionData = predictionData.drop('label', axis=1)

        if silent is False:
            print('classifying: ' + str(marker))

        # shuffle the data
        combined_data = combined_data.sample(frac=1) # shuffle it
        
        # prepare the training data and training labels
        to_train = combined_data.loc[combined_data['label'].isin(['pos', 'neg'])]
        training_data = to_train.drop('label', axis=1)
        training_labels = to_train[['label']]
        trainD = training_data.values
        trainL = training_labels.values
        trainL = [item for sublist in trainL for item in sublist]
        

        #start = time.time()

        # Function for the classifier
        #mlp = MLPClassifier(**kwargs) # CHECK
        #model = GradientBoostingClassifier()
        
        model = HistGradientBoostingClassifier()
        model.fit(trainD, trainL)
        # predict
        pred = model.predict(predictionData.values)
        prob = model.predict_proba(predictionData.values)        
        prob = [item[0] for item in prob]
        
        #end = time.time()
        #print(end - start)
        
        # find the mid point based on the predictions (used for rescaling data later)
        midpoint = find_midpoint(data=predictionData[marker].values, labels=pred)
        
        # return
        return marker, pred, prob, midpoint

    # Run the function
    if silent is False:
        print("Fitting model for classification:")
    r_gatorClassifier = lambda x: gatorClassifier (trainingData_result=x)
    gatorClassifier_result = list(map(r_gatorClassifier, trainingData_result))    
    
    ###########################################################################
    # step-7 : Consolidate the results into a dataframe
    ###########################################################################

    # consolidate results
    markerOrder = []
    for i in range(len(gatorClassifier_result)):
        markerOrder.append(gatorClassifier_result[i][0])
        
    prediction_results = []
    for i in range(len(gatorClassifier_result)):
        prediction_results.append(gatorClassifier_result[i][1])
    prediction_results = pd.DataFrame(prediction_results, index=markerOrder, columns=pre_processed_data.index).T
        
    probability_results = []
    for i in range(len(gatorClassifier_result)):
        probability_results.append(gatorClassifier_result[i][2])
    probability_results = pd.DataFrame(probability_results, index=markerOrder, columns=pre_processed_data.index).T 

    midpoints_dict = {}
    for i in range(len(gatorClassifier_result)):
        midpoints_dict[markerOrder[i]] = gatorClassifier_result[i][3]
    
    ###########################################################################
    # step-8 : Final cleaning of predicted results with UNET results
    ###########################################################################
    
    # bonafide_cells_result_copy = bonafide_cells_result.copy()
    # bonafide_cells_result = bonafide_cells_result[0]
    
    def anomalyDetector (pre_processed_data, bonafide_cells_result, prediction_results):
        # unravel data
        marker = bonafide_cells_result[0]
        pos = bonafide_cells_result[1]
        neg = bonafide_cells_result[2]
        
        if silent is False:
            print("Processing: " + str(marker))
        # prepare data
        X = pre_processed_data.drop(marker, axis=1)
        # scale data
        scaler = StandardScaler()
        X_scaled = scaler.fit_transform(X)
        # model data
        model = LocalOutlierFactor(n_neighbors=20)
        model.fit(X_scaled)
        outlier_scores = model.negative_outlier_factor_
        outliers = pre_processed_data[outlier_scores < -1].index
        
        # common elements betwenn outliers and true neg
        posttoneg = list(set(outliers).intersection(set(neg)))       
        # similarly is there any cells in negative that needs to be relocated to positive?
        negtopos = list(set(pos).intersection(set(prediction_results[prediction_results[marker]=='neg'].index)))
        
        # mutate the prediction results
        prediction_results.loc[posttoneg, marker] = 'neg'
        prediction_results.loc[negtopos, marker] = 'pos'
        
        # results
        results = prediction_results[[marker]]
        return results
    
    # Run the function
    if silent is False:
        print("Running Anomaly Detection")
    r_anomalyDetector = lambda x: anomalyDetector (bonafide_cells_result = x,
                                                   pre_processed_data = pre_processed_data, 
                                                   prediction_results = prediction_results)
    
    # as the Anomaly Detection uses the rest of the data it cannot be run on 1 marker
    if len(bonafide_cells_result) > 1:
        anomalyDetector_result = list(map(r_anomalyDetector, bonafide_cells_result))   
        # final prediction
        prediction_results = pd.concat(anomalyDetector_result, axis=1)
    

    ###########################################################################
    # step-9 : Reorganizing all predictions into a final dataframe
    ###########################################################################
    
    # re introduce failed markers
    if len(failedMarkersinData) > 0 :
        for name in failedMarkersinData:
            prediction_results[name] = 'neg'
            
        # modify the highest value element to be pos
        #modify_prediction_results(rawprocessed, prediction_results, failedMarkersinData)
        
        # identify midpoints for the failed markers (second largest element)
        max_values_dict = get_second_highest_values (rawprocessed, failedMarkersinData)
        
        # update midpoints_dict
        midpoints_dict.update(max_values_dict)
        
        # add the column to pre_processed data for rescaling
        columns_to_concat = rawprocessed[failedMarkersinData]
        pre_processed_data = pd.concat([pre_processed_data, columns_to_concat], axis=1)
        

    ###########################################################################
    # step-10 : Rescale data
    ###########################################################################

    # marker = 'ECAD'
    def rescaleData (marker, pre_processed_data, prediction_results, midpoints_dict):  
        if silent is False:
            print("Processing: " + str(marker))
        # unravel data
        data = pre_processed_data[marker].values
        labels = prediction_results[marker].values
        midpoint = midpoints_dict.get(marker)
        
        # reformat data such that all negs and pos are sorted based on the midpoint
        rescaled = modify_negatives_vectorized(data, 
                                               labels, 
                                               midpoint)
        
        # sigmoid scaling to convert the data between 0-1 based on the midpoint
        if rescaleMethod == 'sigmoid':
            rescaled_data = sigmoid (rescaled, midpoint=midpoint)
        
        if rescaleMethod == 'minmax':
            rescaled_data = scale_data(rescaled, midpoint=midpoint)

        # return
        return rescaled_data
    
    # Run the function
    if silent is False:
        print("Rescaling the raw data")
    r_rescaleData = lambda x: rescaleData (marker=x, 
                                           pre_processed_data=pre_processed_data, 
                                           prediction_results=prediction_results, 
                                           midpoints_dict=midpoints_dict)   
    rescaleData_result = list(map(r_rescaleData, pre_processed_data.columns)) 
    rescaledData = pd.DataFrame(rescaleData_result, index=pre_processed_data.columns, columns=pre_processed_data.index).T

    

    ###########################################################################
    # step-8 : create a new adata object with the results
    ###########################################################################
    
    
    final_markers = pre_processed_data.columns
    intial_markers = rawData.columns
    ordered_final_markers = [marker for marker in intial_markers if marker in final_markers]
    # final raw data
    rd = rawData[ordered_final_markers].reindex(adata.obs.index)
    # final scaled data
    rescaledData = rescaledData[ordered_final_markers].reindex(adata.obs.index)
    # final pre-processed data
    pre_processed_data = pre_processed_data[ordered_final_markers].reindex(adata.obs.index)
    
    
    # reindex prediction results
    prediction_results = prediction_results.reindex(adata.obs.index)
    #probability_results = probability_results.reindex(adata.obs.index)
    
    # create AnnData object
    bdata = ad.AnnData(rd)
    bdata.obs = adata.obs
    bdata.raw = bdata
    bdata.X = rescaledData
    # add the pre-processed data as a layer
    bdata.layers["preProcessed"] = pre_processed_data
    bdata.uns = adata.uns
    
    # save the prediction results in anndata object
    bdata.uns[str(label)] = prediction_results
    #bdata.uns[str(label)] = probability_results
        
    # Save data if requested
    if outputDir is not None:
        finalPath = pathlib.Path(outputDir + '/GATOR/gatorPredict')
        if not os.path.exists(finalPath):
            os.makedirs(finalPath)
        if len(gatorObjectPath) > 1:
            imid = 'gatorPredict'
        else:
            imid = gatorObjectPath[0].stem 
        bdata.write(finalPath / f'{imid}.h5ad')

    return bdata


    
    
    
    
    
    
    
    
    
    